"""
    Test module for cases related to mlflow_plugin
"""

import unittest
from unittest.mock import patch, MagicMock
from mlflow.exceptions import MlflowException, RestException
from ..cogflow.plugins.mlflowplugin import MlflowPlugin
from ..cogflow import (
    delete_registered_model,
    search_registered_models,
    load_model,
    register_model,
    autolog,
    create_registered_model,
    create_model_version,
    set_tracking_uri,
    set_experiment,
    get_artifact_uri,
    start_run,
    end_run,
    log_param,
    log_metric,
    log_model,
    search_model_versions,
)


class TestMlflowPlugin(unittest.TestCase):
    """
    Test Class for cases related to mlflow_plugin
    """

    def setUp(self):
        """
            Initial setup
        :return:
        """
        self.mlflow_plugin = MlflowPlugin()

    @patch("mlflow.get_artifact_uri")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_get_artifact_uri_with_run_id(
        self, mock_plugin_activation, mock_get_artifact_uri
    ):
        """
            test for get_artifact_uri_with_run_id
        :param mock_get_artifact_uri:
        :return:
        """
        # Mocking the mlflow get_artifact_uri method to return a specific URI
        mock_get_artifact_uri.return_value = "s3://your-bucket/artifacts/123"

        result = get_artifact_uri("123")

        # Asserting that the mock method was called with the correct argument
        mock_get_artifact_uri.assert_called_once()

        # Asserting that the method returned the expected artifact URI
        self.assertEqual(result, "s3://your-bucket/artifacts/123")
        mock_plugin_activation.assert_called_once()

    @patch("mlflow.set_experiment")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_set_experiment(self, mock_plugin_activation, mock_set_experiment):
        """
            test for set_experiment
        :param mock_set_experiment:
        :return:
        """
        # Mocking the mlflow set_experiment method to raise an exception
        mock_set_experiment.side_effect = MlflowException("Failed to set experiment")

        with self.assertRaises(MlflowException):
            set_experiment("experiment_name")
        mock_plugin_activation.assert_called_once()

    @patch("mlflow.set_tracking_uri")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_set_tracking_uri(self, mock_plugin_activation, mock_client):
        """
            test for set_tracking_uri
        :param mock_client:
        :return:
        """
        set_tracking_uri("your_tracking_uri")

        mock_client.assert_called_once_with("your_tracking_uri")
        mock_plugin_activation.assert_called_once()

    def test_version(self):
        """
            test for version
        :return:
        """
        version = self.mlflow_plugin.version()

        # Assert that the result matches the mocked version
        self.assertEqual(version, "2.1.1")

    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_is_alive(self, mock_plugin_activation):
        """
            test for is_alive
        :return:
        """
        # Mock the requests.get function to simulate response
        with patch("requests.get") as mock_get:
            mock_response = MagicMock()
            mock_response.status_code = 200  # Assuming Mlflow UI is accessible
            mock_response.text = "OK"  # Mock response message
            mock_get.return_value = mock_response

            mlflow_plugin = MlflowPlugin()  # Create an instance of MlflowPlugin

            status_code, message = mlflow_plugin.is_alive()

            self.assertEqual(status_code, 200)  # Check if status code is 200
            self.assertEqual(message, "OK")  # Check if response message is "OK"
            mock_plugin_activation.assert_called_once()

    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_is_alive_request_not_success(self, mock_plugin_activation):
        """
            test when there is request not successful in is_alive method
        :return:
        """
        # Mock the requests.get function to simulate response
        with patch("requests.get") as mock_get:
            mock_response = MagicMock()
            mock_response.status_code = 404  # Assuming Mlflow UI is not-accessible
            mock_response.text = "NOT-FOUND"  # Mock response message
            mock_get.return_value = mock_response

            mlflow_plugin = MlflowPlugin()  # Create an instance of MlflowPlugin

            status_code, message = mlflow_plugin.is_alive()

            self.assertEqual(status_code, 404)
            self.assertEqual(message, "NOT-FOUND")
            mock_plugin_activation.assert_called_once()

    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_is_alive_request_exception(self, mock_plugin_activation):
        """
            test when there is exception occured in is_alive method
        :return:
        """
        # Mock the requests.get function to simulate response
        with patch("requests.get") as mock_get:
            mock_get.side_effect = Exception("An error occurred .")

            mlflow_plugin = MlflowPlugin()  # Create an instance of MlflowPlugin
            with self.assertRaises(Exception):
                status_code, message = mlflow_plugin.is_alive()
            mock_plugin_activation.assert_called_once()

    @patch("mlflow.tracking.client.MlflowClient.create_model_version")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_create_model_version(self, mock_plugin_activation, mock_client):
        """
            test for create_model_version
        :param mock_client:
        :return:
        """
        mock_client.side_effect = MlflowException("Error occured")
        # # Call the method under test and expect it to raise an MlflowException
        with self.assertRaises(MlflowException):
            result = create_model_version("model_name", "source")
        mock_plugin_activation.assert_called_once()

    @patch("mlflow.tracking.client.MlflowClient")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_create_registered_model_exception(
        self, mock_plugin_activation, mock_client
    ):
        """
            test case when exception occurs in create_registered_model
        :param mock_client:
        :return:
        """
        # Define the exception to be raised
        exception_to_raise = MlflowException("API request failed")

        # Mocking the client method to raise the exception
        mock_client.return_value.create_registered_model.side_effect = (
            exception_to_raise
        )

        self.mlflow_plugin.create_registered_model("model_name_4")
        mock_plugin_activation.assert_called_once()

    @patch("mlflow.tracking.client.MlflowClient.create_registered_model")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_create_registered_model(self, mock_plugin_activation, mock_client):
        """
            test case for create_registered_model
        :param mock_client:
        :return:
        """
        create_registered_model("model_name")
        mock_client.assert_called_once()
        mock_plugin_activation.assert_called_once()

    @patch("mlflow.register_model")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_register_model(self, mock_plugin_activation, mock_register_model):
        """
            test case for register_model
        :param mock_register_model:
        :return:
        """
        # Define inputs
        model = MagicMock()
        model_uri = "my_model_uri"
        register_model(model, model_uri)
        mock_register_model.assert_called_once()
        mock_plugin_activation.assert_called_once()

    @patch("mlflow.sklearn.load_model")
    @patch("mlflow.utils.rest_utils.http_request")
    def test_load_model_exception(self, mock_http_request, mock_load_model):
        """
            test for exception in load_model
        :param mock_http_request:
        :param mock_load_model:
        :return:
        """
        # Define inputs
        model_name = "my_model"
        model_version = 1
        expected_model = MagicMock()
        mock_load_model.return_value = expected_model

        # Mocking the HTTP request to raise MlflowException
        mock_http_request.side_effect = MlflowException(
            "API request failed with exception HTTPConnectionPool: "
            "Max retries exceeded with url: http://127.0.0.1:5001"
        )
        try:
            mock_load_model.assert_not_called()
        except MlflowException:
            with self.assertRaises(MlflowException):
                loaded_model = load_model(model_name, model_version)

    @patch("cogflow.cogflow.plugins.mlflowplugin.MlflowPlugin.load_model")
    def test_load_model(self, mock_load_model):
        """
        Test load_model method in mlflow_plugin.
        """
        # Define inputs
        model_name = "tracking-quickstart"
        model_version = 1
        expected_model = MagicMock()

        # Set the return value of the mocked load_model function
        mock_load_model.return_value = expected_model

        # Call the load_model method
        result = load_model(model_name, model_version)

        # Verify that the result is equal to the expected model
        self.assertEqual(result, expected_model)

    @patch("mlflow.autolog")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_autolog(self, mock_plugin_activation, mock_autolog):
        """
            test for autolog
        :param mock_autolog:
        :return:
        """
        # Call the method under test
        autolog()

        # Assert that autolog was called
        mock_autolog.assert_called_once()
        mock_plugin_activation.assert_called_once()

    @patch("mlflow.tracking.client.MlflowClient.delete_registered_model")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_delete_registered_model(
        self, mock_plugin_activation, mock_delete_registered_model
    ):
        """
            test for delete_registered_model
        :param mock_delete_registered_model:
        :return:
        """
        # Mock successful deletion
        model_name = "test_model"
        mock_delete_registered_model.return_value = True

        # Call the method under test
        result = delete_registered_model(model_name)

        # Assert that the method returns True
        self.assertTrue(result)

        # Assert that delete_registered_model was called with the correct argument
        mock_delete_registered_model.assert_called_once_with(model_name)
        mock_plugin_activation.assert_called_once()

    @patch("mlflow.search_registered_models")
    def test_search_registered_models_exception(self, mock_search_registered_models):
        """
            test for exception occurs when search for registered model
        :param mock_search_registered_models:
        :return:
        """
        # Mock any other unexpected exception
        mock_search_registered_models.side_effect = MlflowException(
            "API request failed with exception HTTPConnectionPool: "
            "Max retries exceeded with url: http://127.0.0.1:5001"
        )
        # Assert that the method raises the expected exception
        try:
            mock_search_registered_models.assert_not_called()
        except MlflowException:
            with self.assertRaises(MlflowException):
                self.mlflow_plugin.search_registered_models()

    @patch("mlflow.tracking.client.MlflowClient.search_registered_models")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_search_registered_models(
        self, mock_plugin_activation, mock_search_registered_models
    ):
        """
            test for search_registered_model
        :param mock_search_registered_models:
        :return:
        """
        mock_search_registered_models.return_value = None
        # Call the method under test
        search_registered_models()

        # Assert that autolog was called
        mock_search_registered_models.assert_called_once()
        mock_plugin_activation.assert_called_once()

    @patch("mlflow.start_run")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_start_run_with_experiment_and_run_name(
        self, mock_plugin_activation, mock_start_run
    ):
        """
            test for start_run_with_experiment with run_name
        :param mock_start_run:
        :return:
        """
        experiment_name = "test_experiment"
        run_name = "test_run"
        start_run(run_name=run_name)
        # Assert that start_run was called with the correct arguments
        mock_start_run.assert_called_once()
        mock_plugin_activation.assert_called_once()

    @patch("mlflow.end_run")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_end_run(self, mock_plugin_activation, mock_end_run):
        """
            test for end_run
        :param mock_end_run:
        :return:
        """
        # Call the method under test
        end_run()
        # Assert that end_run was called
        mock_end_run.assert_called_once()
        mock_plugin_activation.assert_called_once()

    @patch("mlflow.log_param")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_log_param(self, mock_plugin_activation, mock_log_param):
        """
            test for log_param
        :param mock_log_param:
        :return:
        """
        # Define inputs
        run = MagicMock()
        params = {"param1": 10, "param2": "value"}

        # Call the method under test
        log_param(run, params)

        # Assert that log_param was called with the correct arguments
        mock_log_param.assert_called_once()
        mock_plugin_activation.assert_called_once()

    @patch("mlflow.log_metric")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_log_metric(self, mock_plugin_activation, mock_log_metric):
        """
            test for log_metric
        :param mock_log_metric:
        :return:
        """
        # Define inputs
        run = MagicMock()
        metrics = {"accuracy": 0.85, "loss": 0.1}

        # Call the method under test
        log_metric(run, metrics)

        # Assert that log_metric was called with the correct arguments
        mock_log_metric.assert_called_once()
        mock_plugin_activation.assert_called_once()

    @patch("requests.post")
    @patch("os.getenv")
    @patch("mlflow.active_run")
    @patch(
        "cogflow.cogflow.plugins.notebook_plugin.NotebookPlugin.get_model_latest_version"
    )
    @patch("mlflow.sklearn.log_model")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_log_model(
        self,
        mock_plugin_activation,
        mock_log_model,
        mock_model_version,
        mock_active_run,
        mock_env,
        mock_requests_post,
    ):
        """
            test for log_model
        :param mock_log_model:
        :return:
        """
        # Define inputs
        sk_model = MagicMock()
        artifact_path = "model"
        # Define any other necessary inputs for the log_model method

        mock_run = MagicMock()
        mock_run.info.run_id = "12345"

        # Set the return value of mlflow.active_run()
        mock_active_run.return_value = mock_run

        mock_env.side_effect = lambda x: {
            "API_BASEPATH": "http://randomn",
        }[x]

        mock_requests_post.return_value.status_code = 201

        mock_model_version.return_value = 1
        # Call the method under test
        result = log_model(sk_model=sk_model, artifact_path=artifact_path)
        # Assert that log_model was called with the correct arguments
        mock_log_model.assert_called_once()
        mock_plugin_activation.assert_called()

    @patch("requests.post")
    @patch("os.getenv")
    @patch("mlflow.sklearn.log_model")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_log_model_exception(
        self, mock_plugin_activation, mock_log_model, mock_env, mock_requests_post
    ):
        """
            test log_model when exception occurs
        :param mock_log_model:
        :return:
        """
        # Define inputs
        sk_model = MagicMock()
        artifact_path = "model"
        # Define any other necessary inputs for the log_model method

        mock_env.side_effect = lambda x: {
            "API_BASEPATH": "http://randomn",
        }[x]

        mock_requests_post.return_value.status_code = 201

        # Set up the side effect to raise MlflowException
        mock_log_model.side_effect = MlflowException(
            "API request failed with exception HTTPConnectionPool: "
            "Max retries exceeded with url: http://127.0.0.1:5001"
        )

        # Call the method under test and assert that it raises an exception
        with self.assertRaises(MlflowException):
            self.mlflow_plugin.log_model(sk_model=sk_model, artifact_path=artifact_path)
        mock_plugin_activation.assert_called_once()

    @patch("mlflow.MlflowClient.search_model_versions")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_search_model_versions(
        self, mock_verify_activation, mock_search_model_versions
    ):
        filter_string = "name='custom_model'"

        # Call the method being tested
        search_model_versions(
            filter_string=filter_string,
        )
        mock_search_model_versions.assert_called_once()
        mock_verify_activation.assert_called_once()

    @patch("mlflow.tracking.client.MlflowClient.get_model_version")
    def test_get_model_uri(self, mock_client):
        mock_model_version = MagicMock()
        mock_model_version.source = "s3://bucket/model/1"

        # Set up the mock client to return the mock model version
        mock_client.return_value = mock_model_version

        # Call the function
        mfp = MlflowPlugin()
        model_uri = mfp.get_model_uri("my_model", "1")

        # Assertions
        self.assertEqual(model_uri, "s3://bucket/model/1")
        mock_client.assert_called_once_with(name="my_model", version="1")

    @patch("mlflow.tracking.MlflowClient.get_model_version")
    def test_model_version_not_found(self, mock_client):
        # Set up the mock client to raise a RestException
        mock_client.side_effect = RestException(
            json=dict(
                error_code="RESOURCE_DOES_NOT_EXIST", message="Model version not found"
            )
        )

        # Call the function
        mfp = MlflowPlugin()
        with self.assertRaises(RestException):
            mfp.get_model_uri("my_model", "999")

        # Assertions
        mock_client.assert_called_once_with(name="my_model", version="999")


if __name__ == "__main__":
    unittest.main()



"""
This module provides functionality related to Model actions via plugin.
"""

import json
import os
import subprocess
import sys
from datetime import datetime
import joblib
from .mlflowplugin import MlflowPlugin
from .. import plugin_config
from ..pluginmanager import PluginManager
from ..util import custom_serializer
from ..util import make_post_request, make_delete_request, make_get_request
from .kubeflowplugin import KubeflowPlugin


class NotebookPlugin:
    """
    Class for defining reusable components.
    """

    def __init__(self):
        """
        Initializes the ModelPlugin class.
        """
        self.section = "notebook_plugin"

    @staticmethod
    def link_model_to_dataset(dataset_id, model_id):
        """
        Links a model to a dataset using the provided API endpoint.

        This method sends a POST request to the API to associate a specified model
        with a given dataset. It uses the user's ID defined in the plugin configuration.

        Args:
            dataset_id (str): The ID of the dataset to link to the model.
            model_id (str): The ID of the model to be linked to the dataset.

        Returns:
            Response: The response object from the API call.

        Raises:
            requests.exceptions.RequestException: An error occurred when making the POST request.
        """
        # Verify plugin activation
        PluginManager().verify_activation(NotebookPlugin().section)

        PluginManager().load_config()

        data = {
            "user_id": plugin_config.JUPYTER_USER_ID,
            "model_id": model_id,
            "dataset_id": dataset_id,
        }
        # call the api
        url = os.getenv(plugin_config.API_BASEPATH) + "/link_dataset_model"
        return make_post_request(url, data=data)

    def save_model_details_to_db(self, registered_model_name):
        """
        store model details in database
        :param registered_model_name: name of the registered model
        :return: id of model
        """
        # Verify plugin activation
        PluginManager().verify_activation(NotebookPlugin().section)

        PluginManager().load_config()

        data = {
            "name": registered_model_name,
            "version": self.get_model_latest_version(registered_model_name),
            "type": "sklearn",
            "user_id": plugin_config.JUPYTER_USER_ID,
        }

        # call the api to register model
        url = os.getenv(plugin_config.API_BASEPATH) + "/models"
        return make_post_request(url, data=data)

    @staticmethod
    def get_model_latest_version(registered_model_name: str):
        """
        return the latest version of registered model
        :param registered_model_name: model name to get the versions
        :return: latest version
        """
        # Verify plugin activation
        PluginManager().verify_activation(NotebookPlugin().section)

        latest_version_info = MlflowPlugin().search_model_versions(
            filter_string=f"name='{registered_model_name}'"
        )
        sorted_model_versions = sorted(
            latest_version_info, key=lambda x: int(x.version), reverse=True
        )

        if sorted_model_versions:
            latest_version = sorted_model_versions[0]
            print("Latest Version:", latest_version.version)
            print("Status:", latest_version.status)
            print("Stage:", latest_version.current_stage)
            print("Description:", latest_version.description)
            print("Last Updated:", latest_version.last_updated_timestamp)
            return latest_version.version

        print(f"No model versions found for {registered_model_name}")
        return 1

    @staticmethod
    def save_model_uri_to_db(model_id, model_uri):
        """
            method to call the api to save model uri
        :param model_id: model id of the model
        :param model_uri: model uri
        :return: API response
        """
        # Verify plugin activation
        PluginManager().verify_activation(NotebookPlugin().section)

        PluginManager().load_config()

        # call the api for saving model_uri
        data = {
            "user_id": plugin_config.JUPYTER_USER_ID,
            "model_id": model_id,
            "uri": model_uri,
            "description": f"model uri of model id :{model_id}",
        }
        url = os.getenv(plugin_config.API_BASEPATH) + "/models/uri"
        return make_post_request(url, data=data)

    @staticmethod
    def delete_pipeline_details_from_db(pipeline_id):
        """
        delete the pipeline details
        :param pipeline_id: pipeline id
        :return:
        """
        # Verify plugin activation
        PluginManager().verify_activation(NotebookPlugin().section)

        PluginManager().load_config()

        url = os.getenv(plugin_config.API_BASEPATH) + "/pipeline"
        return make_delete_request(url=url, path_params=pipeline_id)

    @staticmethod
    def list_runs_by_pipeline_id(pipeline_id):
        """
        list the pipeline run details
        :param pipeline_id: pipeline_id
        :return: list of run details
        """
        # Verify plugin activation
        PluginManager().verify_activation(NotebookPlugin().section)

        PluginManager().load_config()

        url = os.getenv(plugin_config.API_BASEPATH) + "/pipeline/runs"
        return make_get_request(url=url, path_params=pipeline_id)

    @staticmethod
    def delete_run_details_from_db(pipeline_id):
        """
         delete the pipeline details
        :param pipeline_id: pipeline_id
        :return: successful deletion message or 404 error
        """
        # Verify plugin activation
        PluginManager().verify_activation(NotebookPlugin().section)

        PluginManager().load_config()

        url = os.getenv(plugin_config.API_BASEPATH) + "/pipeline/runs"
        return make_delete_request(url=url, path_params=pipeline_id)

    @staticmethod
    def get_pipeline_id_by_name(pipeline_name):
        """
        Retrieves the pipeline ID for a given pipeline name.

        Args:
            pipeline_name (str): The name of the pipeline to fetch the ID for.

        Returns:
            str: The ID of the specified pipeline if found.

        Raises:
            ValueError: If no pipeline with the given name is found.

        Example:
            pipeline_id = NotebookPlugin.get_pipeline_id_by_name("Example Pipeline")
        """
        kfp = KubeflowPlugin()
        pipelines_response = kfp.client().list_pipelines()
        pipeline_id = None
        if pipelines_response.pipelines:
            for pipeline in pipelines_response.pipelines:
                if pipeline.name == pipeline_name:
                    pipeline_id = pipeline.id
                    return pipeline_id

        if not pipeline_id:
            print(f"No pipeline found with the name '{pipeline_name}'")

    @staticmethod
    def list_pipelines_by_name(pipeline_name):
        """
        Lists all versions and runs of the specified pipeline by name.

        Args:
            pipeline_name (str): The name of the pipeline to fetch details for.

        Returns:
            dict: A dictionary containing the pipeline ID, versions,
             and runs of the specified pipeline.

        Raises:
            ValueError: If the pipeline name is invalid or not found.
            Exception: For any other issues encountered during the fetch operations.
        """
        # Fetch all versions of the specified pipeline
        kfp = KubeflowPlugin()

        pipeline_id = NotebookPlugin.get_pipeline_id_by_name(pipeline_name)
        versions_response = kfp.list_pipeline_versions(pipeline_id=pipeline_id)
        run_list = NotebookPlugin.list_runs_by_pipeline_id(pipeline_id=pipeline_id)
        result_dict = {
            "pipeline_id": pipeline_id,
            "versions": versions_response.versions,
            "runs": run_list,
        }
        return result_dict

    @staticmethod
    def save_pipeline_details_to_db(details):
        """
            save the details related to pipeline to the database
        :param details: dictionary with all the details of pipeline,run_details,task_details,experiments
        :return:
        """
        # Verify plugin activation
        PluginManager().verify_activation(NotebookPlugin().section)

        PluginManager().load_config()

        data = json.dumps(details, default=custom_serializer, indent=4)
        url = os.getenv(plugin_config.API_BASEPATH) + "/pipeline/add"
        make_post_request(url=url, data=data)

    def log_model_by_model_file(self, model_file_path, model_name):
        """
            log_model in cogflow with the model_file
        :param model_file_path: file_path of model
        :param model_name: name of the model
        :return:
            data = {
                "artifact_uri" : 'artifact_uri of the model',
                "version" : "model version"
            }
        """
        model = self.load_pkl(model_file_path)
        mfp = MlflowPlugin()
        mfp.mlflow.set_experiment(f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}_exp")
        with mfp.mlflow.start_run() as run:
            run_id = run.info.run_id
            mfp.mlflow.sklearn.log_model(
                model, "model", registered_model_name=model_name
            )
            print("Artifact_path", run.info.artifact_uri)
            print("run_id", run_id)
            print("model_name", model_name)
            latest_version = self.get_model_latest_version(model_name)
            data = {
                "artifact_uri": f"{run.info.artifact_uri}/model",
                "version": latest_version,
            }
            return data

    def install_and_import(self, package):
        """
            install and import the given package
        :param package: package to be installed
        :return:
        """
        try:
            __import__(package)
        except ImportError:
            subprocess.check_call([sys.executable, "-m", "pip", "install", package])
            __import__(package)

    def load_pkl(self, file_path):
        """
            load the pkl file to joblib
        :param file_path: path of .pkl file
        :return:
        """
        try:
            with open(file_path, "rb") as file:
                return joblib.load(file)
        except ModuleNotFoundError as e:
            missing_module = str(e).split("'")[1]
            print(f"Module {missing_module} not found. Installing...")
            self.install_and_import(missing_module)
            print(f"Module {missing_module} installed. Trying to load the file again.")
            return self.load_pkl(file_path)

    def deploy_model(self, model_name, model_version, isvc_name):
        """

        :param model_name: name of the model
        :param model_version: version of the model
        :param isvc_name: service name to be created for the deployed model
        :return:
        """
        try:
            PluginManager().verify_activation(NotebookPlugin().section)
            mfp = MlflowPlugin()
            model_uri = mfp.get_model_uri(model_name=model_name, version=model_version)
            kfp = KubeflowPlugin()
            kfp.serve_model_v1(model_uri, name=isvc_name)
            return {
                "status": True,
                "msg": f"Model {model_name} deployed with service {isvc_name}",
            }
        except Exception as exp:
            raise exp

    @staticmethod
    def run_kubectl_command(command):
        """
        Run a kubectl command and return the output.

        Args:
            command (list): The kubectl command to run.

        Returns:
            str: The output from the kubectl command, or None if the command failed.
        """
        try:
            print(f"Running command: {' '.join(command)}")
            result = subprocess.run(command, capture_output=True, text=True, check=True)
            print(f"Command output: {result.stdout}")
            return result.stdout
        except subprocess.CalledProcessError as e:
            print(f"Command '{' '.join(command)}' failed with error: {e.stderr}")
            return None
        except Exception as e:
            print(f"Unexpected error running command '{' '.join(command)}': {str(e)}")
            return None

    @staticmethod
    def get_pods_in_namespace(namespace):
        """
        Get the list of pods in a given namespace.

        Args:
            namespace (str): The Kubernetes namespace to query.

        Returns:
            list: A list of pod items in the given namespace, or an empty list if none are found.
        """
        command = ["kubectl", "get", "pods", "-n", namespace, "-o", "json"]
        output = NotebookPlugin.run_kubectl_command(command)
        if output:
            pod_info = json.loads(output)
            return pod_info["items"]
        return []

    @staticmethod
    def get_pod_logs(namespace, pod_name, container_name=None):
        """
        Get the logs for a specific pod.

        Args:
            namespace (str): The Kubernetes namespace.
            pod_name (str): The name of the pod.
            container_name (str, optional): The name of the container in the pod. Defaults to None.

        Returns:
            str: The logs from the pod, or None if the command failed.
        """
        command = ["kubectl", "logs", "-n", namespace, pod_name]
        if container_name:
            command.append(container_name)
        logs = NotebookPlugin.run_kubectl_command(command)
        return logs

    @staticmethod
    def get_pod_prefix(inference_service_name):
        """
        Get the pod prefix for a given inference service.

        Args:
            inference_service_name (str): The name of the inference service.

        Returns:
            str: The latest ready revision of the inference service, or None if not found.
        """
        command = [
            "kubectl",
            "get",
            "inferenceservice",
            inference_service_name,
            "-o",
            "json",
        ]
        output = NotebookPlugin.run_kubectl_command(command)
        if output:
            isvc_info = json.loads(output)
            # logger.info(f"InferenceService JSON: {json.dumps(isvc_info, indent=2)}")
            # Adjust the key based on the actual JSON structure
            if (
                "status" in isvc_info
                and "components" in isvc_info["status"]
                and "predictor" in isvc_info["status"]["components"]
            ):
                predictor_info = isvc_info["status"]["components"]["predictor"]
                if "latestReadyRevision" in predictor_info:
                    latest_ready_revision = predictor_info["latestReadyRevision"]
                    return latest_ready_revision
            return "Key 'latestReadyRevision' not found in InferenceService status"
        return None

    @staticmethod
    def get_logs_for_inference_service(namespace, inference_service_name):
        """
        Get the logs for all pods related to a specific inference service.

        Args:
            namespace (str): The Kubernetes namespace.
            inference_service_name (str): The name of the inference service.

        Returns:
            dict: A dictionary containing the logs for each related pod, or None if an error occurred.
        """
        pod_prefix = NotebookPlugin.get_pod_prefix(inference_service_name)
        if not pod_prefix:
            return f"Failed to retrieve pod prefix for inference service '{inference_service_name}'"
        pods = NotebookPlugin.get_pods_in_namespace(namespace)
        if not pods:
            return f"No pods found in namespace '{namespace}'"
        related_pods = [pod for pod in pods if pod_prefix in pod["metadata"]["name"]]
        if not related_pods:
            return f"No pods found for inference service '{inference_service_name}' with prefix '{pod_prefix}'"
        logs_output = {}
        for pod in related_pods:
            pod_name = pod["metadata"]["name"]
            print(f"Retrieving logs for pod: {pod_name} in namespace: {namespace}")
            logs = NotebookPlugin.get_pod_logs(namespace, pod_name)
            if logs:
                logs_output[pod_name] = logs
            else:
                print(f"Failed to get logs for pod  '{pod_name}'")
        return logs_output



"""
This module defines constants for configuring Mlflow and AWS S3, as well as database credentials
 and MinIO settings.

Attributes:
    TRACKING_URI (str): The URI of the Mlflow tracking server. Used to specify where Mlflow logs and
    artifacts should be stored.
    S3_ENDPOINT_URL (str): The endpoint URL for the AWS S3 service. Needed for accessing and
     storing data in an S3 bucket.
    ACCESS_KEY_ID (str): The access key ID for AWS S3 authentication. Used in conjunction with
     SECRET_ACCESS_KEY for secure access to S3.
    SECRET_ACCESS_KEY (str): The secret access key for AWS S3 authentication. Paired with
     ACCESS_KEY_ID for secure access to S3.
    TIMER_IN_SEC (int): The interval in seconds for operations that require timing, such as
     periodic checks or updates. Default is set to 10 seconds.
    ML_TOOL (str): The name of the machine learning tool. Currently set to "mlflow" to
     specify the use of the Mlflow framework.

    ML_DB_USERNAME (str): Username for connecting to the machine learning database (ML_DB).
     Essential for database access authentication.
    ML_DB_PASSWORD (str): Password for connecting to the machine learning database (ML_DB).
     Should be kept secure.
    ML_DB_HOST (str): Host address for connecting to the machine learning database (ML_DB).
    ML_DB_PORT (str): Port number for connecting to the machine learning database (ML_DB).
    ML_DB_NAME (str): Name of the machine learning database (ML_DB).

    Cogflow_DB_USERNAME (str): Username for connecting to the CogFlow database. Important for
     authenticating access to the database.
    Cogflow_DB_PASSWORD (str): Password for connecting to the CogFlow database.
     Should be stored securely.
    Cogflow_DB_HOST (str): Host address for connecting to the CogFlow database.
    Cogflow_DB_PORT (str): Port number for connecting to the CogFlow database.
    Cogflow_DB_NAME (str): Name of the CogFlow database.

    MINIO_ENDPOINT_URL (str): The endpoint URL for the MinIO service, an alternative to AWS S3.
     Used for accessing and storing data.
    MINIO_ACCESS_KEY (str): The access key for MinIO authentication. Used with
     MINIO_SECRET_ACCESS_KEY for secure access to MinIO.
    MINIO_SECRET_ACCESS_KEY (str): The secret access key for MinIO authentication. Combined with
     MINIO_ACCESS_KEY for secure access.
"""

COGFLOW_CONFIG_FILE_PATH = "COGFLOW_CONFIG_FILE_PATH"

TRACKING_URI = "MLFLOW_TRACKING_URI"
S3_ENDPOINT_URL = "MLFLOW_S3_ENDPOINT_URL"
ACCESS_KEY_ID = "AWS_ACCESS_KEY_ID"
SECRET_ACCESS_KEY = "AWS_SECRET_ACCESS_KEY"
TIMER_IN_SEC = 10
ML_TOOL = "mlflow"

ML_DB_USERNAME = "ML_DB_USERNAME"
ML_DB_PASSWORD = "ML_DB_PASSWORD"
ML_DB_HOST = "ML_DB_HOST"
ML_DB_PORT = "ML_DB_PORT"
ML_DB_NAME = "ML_DB_NAME"

COGFLOW_DB_USERNAME = "DB_USER"
COGFLOW_DB_PASSWORD = "DB_PASSWORD"
COGFLOW_DB_HOST = "DB_HOST"
COGFLOW_DB_PORT = "DB_PORT"
COGFLOW_DB_NAME = "DB_NAME"

MINIO_ENDPOINT_URL = "MLFLOW_S3_ENDPOINT_URL"
MINIO_ACCESS_KEY = "AWS_ACCESS_KEY_ID"
MINIO_SECRET_ACCESS_KEY = "AWS_SECRET_ACCESS_KEY"

SQLALCHEMY_TEST_DATABASE_URI = "sqlite:///test.db"
TESTING_CONFIG = "config.app_config.TestingConfig"

JUPYTER_USER_ID = 0

API_BASEPATH = "API_BASEPATH"



import os
import unittest
from unittest.mock import MagicMock, patch

from kubernetes.client.models import V1EnvVar
from ..cogflow import (
    plugin_config,
    pipeline,
    create_component_from_func,
    client,
    input_path,
    output_path,
    serve_model_v2,
    serve_model_v1,
    get_model_url,
    load_component,
    delete_pipeline,
)
from ..cogflow.plugins.kubeflowplugin import KubeflowPlugin, CogContainer


class TestKubeflowPlugin(unittest.TestCase):
    def setUp(self):
        self.kfp_plugin = KubeflowPlugin()

    @patch("kfp.dsl.pipeline")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_pipeline_with_name_and_description(
        self, mock_plugin_activation, mock_pipeline
    ):
        pipeline()
        mock_pipeline.assert_called_once()
        mock_plugin_activation.assert_called_once()

    @patch("kfp.components.create_component_from_func")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_create_component_from_func(
        self, mock_plugin_activation, mock_create_component
    ):
        func = MagicMock()
        create_component_from_func(func)
        mock_create_component.assert_called_once()
        mock_plugin_activation.assert_called_once()

    @patch("kfp.Client")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_client(self, mock_plugin_activation, mock_client):
        # Arrange
        client()

        # Assertion
        mock_client.assert_called_once()
        mock_plugin_activation.assert_called_once()

    @patch("kfp.components.load_component_from_url")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_load_component_from_url_success(
        self, mock_plugin_activation, mock_load_component
    ):
        # Mock a successful component loading
        expected_component_spec = MagicMock()
        mock_load_component.return_value = expected_component_spec

        # Define a sample URL
        url = "http://example.com/component.tar.gz"

        # Call the function under test
        result = load_component(url=url)

        # Assert that the function returns the expected component specification
        self.assertEqual(result, expected_component_spec)

        # Assert that load_component_from_url was called with the correct URL
        mock_load_component.assert_called_once_with(url)
        mock_plugin_activation.assert_called_once()

    @patch("kfp.components.InputPath")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_input_path(self, mock_plugin_activation, mock_input):
        # Define a sample label for the input path
        label = "input_data"

        input_path_val = input_path(label)
        mock_input.assert_called_once()
        mock_plugin_activation.assert_called_once()

    @patch("kfp.components.OutputPath")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_output_path(self, mock_plugin_activation, mock_output):
        # Define a sample label for the input path
        label = "output_data"

        input_path = output_path(label)
        mock_output.assert_called_once()
        mock_plugin_activation.assert_called_once()

    @patch.dict(
        os.environ,
        {
            "TRACKING_URI": "tracking_uri_value",
            "S3_ENDPOINT_URL": "s3_endpoint_url_value",
            "ACCESS_KEY_ID": "access_key_id_value",
            "SECRET_ACCESS_KEY": "secret_access_key_value",
            "ADDITIONAL_VAR_1": "additional_value_1",
            "ADDITIONAL_VAR_2": "additional_value_2",
        },
    )
    def test_add_model_access(self):
        # Create an instance of CogContainer
        container = CogContainer()

        # Mock plugin_config to use the environment variable names directly
        with patch.object(plugin_config, "TRACKING_URI", "TRACKING_URI"), patch.object(
            plugin_config, "S3_ENDPOINT_URL", "S3_ENDPOINT_URL"
        ), patch.object(plugin_config, "ACCESS_KEY_ID", "ACCESS_KEY_ID"), patch.object(
            plugin_config, "SECRET_ACCESS_KEY", "SECRET_ACCESS_KEY"
        ):
            # Call the add_model_access method
            container_with_env_vars = container.add_model_access()

            # Assert that the returned value is an instance of CogContainer
            self.assertIsInstance(container_with_env_vars, CogContainer)

            # Expected environment variables explicitly listed in the method
            expected_env_vars = [
                V1EnvVar(name="TRACKING_URI", value="tracking_uri_value"),
                V1EnvVar(name="S3_ENDPOINT_URL", value="s3_endpoint_url_value"),
                V1EnvVar(name="ACCESS_KEY_ID", value="access_key_id_value"),
                V1EnvVar(name="SECRET_ACCESS_KEY", value="secret_access_key_value"),
            ]

            # Expected additional environment variables from os.environ
            additional_env_vars = [
                V1EnvVar(name="ADDITIONAL_VAR_1", value="additional_value_1"),
                V1EnvVar(name="ADDITIONAL_VAR_2", value="additional_value_2"),
            ]

            all_expected_env_vars = expected_env_vars + additional_env_vars

            # Assert that all expected environment variables are added correctly
            for expected_env_var in all_expected_env_vars:
                self.assertIn(expected_env_var, container_with_env_vars.env)

    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_serve_model_v2(self, mock_plugin_activation):
        # Patch Kubernetes client to avoid loading kube config
        with patch("kubernetes.config.load_kube_config"):
            model_uri = "sample_model_uri"
            name = "test_model_name"

            # Call the function and assert that it raises MaxRetryError
            with self.assertRaises(Exception):
                serve_model_v2(model_uri, name)
            mock_plugin_activation.assert_called_once()

    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_serve_model_v2_no_name(self, mock_plugin_activation):
        with patch("kubernetes.config.load_kube_config"):
            model_uri = "sample_model_uri"

            # Call the function and assert that it raises MaxRetryError
            with self.assertRaises(Exception):
                self.kfp_plugin.serve_model_v2(model_uri)
            mock_plugin_activation.assert_called_once()

    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_serve_model_v1_with_exception(self, mock_plugin_activation):
        # Define input parameters
        model_uri = "example_model_uri"
        name = "test_model_name"

        # Call the function and assert that it raises MaxRetryError
        with self.assertRaises(Exception):
            serve_model_v1(model_uri, name)
        mock_plugin_activation.assert_called_once()

    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_get_model_url(self, mock_plugin_activation):
        model_name = "test_model"

        with self.assertRaises(Exception):
            # Call the method you're testing here
            get_model_url(model_name)
        mock_plugin_activation.assert_called_once()

    @patch("requests.delete")
    @patch("requests.get")
    @patch("os.getenv")
    @patch("cogflow.cogflow.plugins.kubeflowplugin.KubeflowPlugin.client")
    def test_delete_pipeline(
        self, mock_client, mock_env, mock_request_get, mock_request_delete
    ):
        # Arrange
        mock_env.side_effect = lambda x: {
            "API_BASEPATH": "http://randomn",
            "COGFLOW_CONFIG_FILE_PATH": "/path",
        }[x]
        mock_request_get.return_value.status_code = 200
        mock_request_delete.return_value.status_code = 200
        plugin = KubeflowPlugin()
        mock_client_instance = mock_client.return_value
        pipeline_id = "test_pipeline_id"

        # Act
        delete_pipeline(pipeline_id)

        # Assert
        mock_client_instance.delete_pipeline.assert_called_once_with(
            pipeline_id=pipeline_id
        )

    @patch("cogflow.cogflow.plugins.kubeflowplugin.KubeflowPlugin.client")
    def test_list_pipeline_versions(self, mock_client):
        # Arrange
        plugin = KubeflowPlugin()
        mock_client_instance = mock_client.return_value
        pipeline_id = "test_pipeline_id"
        expected_response = "expected_response"
        mock_client_instance.list_pipeline_versions.return_value = expected_response

        # Act
        response = plugin.list_pipeline_versions(pipeline_id)

        # Assert
        mock_client_instance.list_pipeline_versions.assert_called_once_with(
            pipeline_id=pipeline_id
        )
        self.assertEqual(response, expected_response)

    @patch("cogflow.cogflow.plugins.kubeflowplugin.KubeflowPlugin.client")
    def test_delete_pipeline_version(self, mock_client):
        # Arrange
        plugin = KubeflowPlugin()
        mock_client_instance = mock_client.return_value
        version_id = "test_version_id"

        # Act
        plugin.delete_pipeline_version(version_id)

        # Assert
        mock_client_instance.delete_pipeline_version.assert_called_once_with(
            version_id=version_id
        )

    @patch("cogflow.cogflow.plugins.kubeflowplugin.KubeflowPlugin.client")
    def test_delete_runs(self, mock_client):
        # Arrange
        plugin = KubeflowPlugin()
        mock_client_instance = mock_client.return_value
        mock_client_instance.runs = MagicMock()
        run_ids = [1, 2]

        # Act
        plugin.delete_runs(run_ids)

        # Assert
        calls = [unittest.mock.call(id=1), unittest.mock.call(id=2)]
        mock_client_instance.runs.delete_run.assert_has_calls(calls, any_order=True)
        self.assertEqual(mock_client_instance.runs.delete_run.call_count, 2)


if __name__ == "__main__":
    unittest.main()



"""
This module provides functionality related to Kubeflow Pipelines.
"""

import os
import time
from datetime import datetime
from typing import Optional, Dict, Any
import kfp
from kfp import dsl, v2
from kserve import (
    KServeClient,
    V1beta1InferenceService,
    V1beta1InferenceServiceSpec,
    V1beta1ModelFormat,
    V1beta1ModelSpec,
    V1beta1PredictorSpec,
    V1beta1SKLearnSpec,
    constants,
    utils,
)
from kubernetes import client
from kubernetes.client import V1ObjectMeta
from kubernetes.client.models import V1EnvVar
from tenacity import retry, wait_exponential, stop_after_attempt

from .. import plugin_config
from ..pluginmanager import PluginManager


class CogContainer(kfp.dsl._container_op.Container):
    """
    Subclass of Container to add model access environment variables.
    """

    def __init__(self, name=None, image=None, command=None, args=None, **kwargs):
        """
        Initializes the CogContainer class.
        """
        super().__init__(name=name, image=image, command=command, args=args, **kwargs)

    def __call__(self, *args, **kwargs):
        """
        Calls the parent class's __call__ method to initialize the container operation,
        and then applies additional model access configurations.

        This method first invokes the base class's __call__ method to create and configure
        the container operation. Afterward, it applies model access environment variables
        by calling `add_model_access()` on the current instance.

        Args:
            *args: Positional arguments to be passed to the base class's __call__ method.
            **kwargs: Keyword arguments to be passed to the base class's __call__ method.

        Returns:
            CogContainer: The instance of CogContainer with model access environment variables
            added. This is the modified container operation that includes additional configurations.
        """
        base_result = super().__call__(*args, **kwargs)

        return base_result.add_model_access()

    def add_model_access(self):
        """
        Adds model access environment variables to the container.

        Returns:
            CogContainer: Container instance with added environment variables.
        """
        # Verify plugin activation
        PluginManager().verify_activation(KubeflowPlugin().section)

        # Adding environment variables
        for key, value in os.environ.items():
            self.add_env_variable(V1EnvVar(name=key, value=value))
        return self


class KubeflowPlugin:
    """
    Class for defining reusable components.
    """

    def __init__(self, image=None, command=None, args=None):
        """
        Initializes the KubeflowPlugin class.
        """
        self.kfp = kfp
        self.kfp.dsl._container_op.Container.AddModelAccess = (
            CogContainer.add_model_access
        )
        self.kfp.dsl._container_op.ContainerOp.AddModelAccess = (
            CogContainer.add_model_access
        )
        self.config_file_path = os.getenv(plugin_config.COGFLOW_CONFIG_FILE_PATH)
        self.v2 = kfp.v2
        self.section = "kubeflow_plugin"

    @staticmethod
    def pipeline(name=None, description=None):
        """
        Decorator function to define Kubeflow Pipelines.

        Args:
            name (str, optional): Name of the pipeline. Defaults to None.
            description (str, optional): Description of the pipeline. Defaults to None.

        Returns:
            Callable: Decorator for defining Kubeflow Pipelines.
        """
        # Verify plugin activation
        PluginManager().verify_activation(KubeflowPlugin().section)

        return dsl.pipeline(name=name, description=description)

    def create_component_from_func(
            self,
            func,
            output_component_file=None,
            base_image=None,
            packages_to_install=None,
    ):
        """
        Create a component from a Python function.

        Args:
            func (Callable): Python function to convert into a component.
            output_component_file (str, optional): Path to save the component YAML file. Defaults
            to None.
            base_image (str, optional): Base Docker image for the component. Defaults to None.
            packages_to_install (List[str], optional): List of additional Python packages
            to install in the component.
            Defaults to None.

        Returns:
            kfp.components.ComponentSpec: Component specification.
        """
        # Verify plugin activation
        PluginManager().verify_activation(KubeflowPlugin().section)

        training_var = kfp.components.create_component_from_func(
            func=func,
            output_component_file=output_component_file,
            base_image=base_image,
            packages_to_install=packages_to_install,
        )
        training_var.add_model_access = CogContainer().add_model_access
        training_var.add_model_access()

        return training_var

    @staticmethod
    def client():
        """
        Get the Kubeflow Pipeline client.

        Returns:
            kfp.Client: Kubeflow Pipeline client instance.
        """
        # Verify plugin activation
        PluginManager().verify_activation(KubeflowPlugin().section)

        return kfp.Client()

    @staticmethod
    def load_component_from_url(url):
        """
        Load a component from a URL.

        Args:
            url (str): URL to load the component from.

        Returns:
            kfp.components.ComponentSpec: Loaded component specification.
        """
        # Verify plugin activation
        PluginManager().verify_activation(KubeflowPlugin().section)

        return kfp.components.load_component_from_url(url)

    @staticmethod
    def input_path(label: str):
        """
        Create an InputPath component.

        Args:
            label (str): Label for the input path.

        Returns:
            InputPath: InputPath component instance.
        """
        # Verify plugin activation
        PluginManager().verify_activation(KubeflowPlugin().section)

        return kfp.components.InputPath(label)

    @staticmethod
    def output_path(label: str):
        """
        Create an OutputPath component.

        Args:
            label (str): Label for the output path.

        Returns:
            OutputPath: OutputPath component instance.
        """
        # Verify plugin activation
        PluginManager().verify_activation(KubeflowPlugin().section)

        return kfp.components.OutputPath(label)

    @staticmethod
    def serve_model_v2(model_uri: str, name: str = None):
        """
        Create a kserve instance.

        Args:
            model_uri (str): URI of the model.
            name (str, optional): Name of the kserve instance. If not provided,
            a default name will be generated.

        Returns:
            None
        """
        # Verify plugin activation
        PluginManager().verify_activation(KubeflowPlugin().section)

        namespace = utils.get_default_target_namespace()
        if name is None:
            now = datetime.now()
            date = now.strftime("%d%M")
            name = f"predictor_model{date}"
        isvc_name = name
        predictor = V1beta1PredictorSpec(
            service_account_name="kserve-controller-s3",
            min_replicas=1,
            model=V1beta1ModelSpec(
                model_format=V1beta1ModelFormat(
                    name=plugin_config.ML_TOOL,
                ),
                storage_uri=model_uri,
                protocol_version="v2",
            ),
        )

        isvc = V1beta1InferenceService(
            api_version=constants.KSERVE_V1BETA1,
            kind=constants.KSERVE_KIND,
            metadata=client.V1ObjectMeta(
                name=isvc_name,
                namespace=namespace,
                annotations={"sidecar.istio.io/inject": "false"},
            ),
            spec=V1beta1InferenceServiceSpec(predictor=predictor),
        )
        kserve = KServeClient()
        kserve.create(isvc)
        time.sleep(plugin_config.TIMER_IN_SEC)

    @staticmethod
    def serve_model_v1(model_uri: str, name: str = None):
        """
        Create a kserve instance version1.

        Args:
            model_uri (str): URI of the model.
            name (str, optional): Name of the kserve instance. If not provided,
            a default name will be generated.

        Returns:
            None
        """
        # Verify plugin activation
        PluginManager().verify_activation(KubeflowPlugin().section)

        isvc_name = name
        namespace = utils.get_default_target_namespace()
        isvc = V1beta1InferenceService(
            api_version=constants.KSERVE_V1BETA1,
            kind=constants.KSERVE_KIND,
            metadata=V1ObjectMeta(
                name=isvc_name,
                namespace=namespace,
                annotations={"sidecar.istio.io/inject": "false"},
            ),
            spec=V1beta1InferenceServiceSpec(
                predictor=V1beta1PredictorSpec(
                    service_account_name="kserve-controller-s3",
                    sklearn=V1beta1SKLearnSpec(storage_uri=model_uri),
                )
            ),
        )

        kclient = KServeClient()
        kclient.create(isvc)
        print("model_uri", model_uri)
        print("isvc", isvc)
        time.sleep(plugin_config.TIMER_IN_SEC)

    @staticmethod
    def get_model_url(model_name: str):
        """
        Retrieve the URL of a deployed model.

        Args:
            model_name (str): Name of the deployed model.

        Returns:
            str: URL of the deployed model.
        """
        # Verify plugin activation
        PluginManager().verify_activation(KubeflowPlugin().section)

        kclient = KServeClient()

        @retry(
            wait=wait_exponential(multiplier=2, min=1, max=10),
            stop=stop_after_attempt(30),
            reraise=True,
        )
        def assert_isvc_created(kserve_client, isvc_name):
            """Wait for the Inference Service to be created successfully."""
            assert kserve_client.is_isvc_ready(
                isvc_name
            ), f"Failed to create Inference Service {isvc_name}."

        assert_isvc_created(kclient, model_name)

        isvc_resp = kclient.get(model_name)
        isvc_url = isvc_resp["status"]["address"]["url"]
        return isvc_url

    @staticmethod
    def delete_served_model(model_name: str):
        """
        Delete a deployed model by its ISVC name.

        Args:
            model_name (str): Name of the deployed model.

        Returns:
            None
        """
        # Verify plugin activation
        PluginManager().verify_activation(KubeflowPlugin().section)

        kclient = KServeClient()

        # Attempt to delete the inference service
        response = kclient.delete(model_name)
        if (
                response.get("status", {}).get("conditions", [{}])[0].get("type")
                == "IngressReady"
        ):
            return f"Inference Service {model_name} has been deleted successfully."
        raise Exception(f"Failed to delete Inference Service {model_name}.")

    @staticmethod
    def load_component_from_file(file_path):
        """
        Load a component from a File.

        Args:
            file_path (str): file_path to load the component from file.

        Returns:
            kfp.components.ComponentSpec: Loaded component specification.
        """
        # Verify plugin activation
        PluginManager().verify_activation(KubeflowPlugin().section)
        return kfp.components.load_component_from_file(file_path)

    @staticmethod
    def load_component_from_text(text):
        """
        Load a component from the text.

        Args:
            text (str):  load the component from text.

        Returns:
            kfp.components.ComponentSpec: Loaded component specification.
        """
        # Verify plugin activation
        PluginManager().verify_activation(KubeflowPlugin().section)
        return kfp.components.load_component_from_text(text)

    def create_run_from_pipeline_func(
            self,
            pipeline_func: ...,
            arguments: Optional[Dict[str, Any]] = None,
            run_name: Optional[str] = None,
            experiment_name: Optional[str] = None,
            namespace: Optional[str] = None,
            pipeline_root: Optional[str] = None,
            enable_caching: Optional[bool] = None,
            service_account: Optional[str] = None,
    ):
        """
            method to create a run from pipeline function
        :param pipeline_func:
        :param arguments:
        :param run_name:
        :param experiment_name:
        :param namespace:
        :param pipeline_root:
        :param enable_caching:
        :param service_account:
        :param experiment_id:
        :return:
        """
        run_details = self.client().create_run_from_pipeline_func(
            pipeline_func,
            arguments,
            run_name,
            experiment_name,
            namespace,
            pipeline_root,
            enable_caching,
            service_account,
        )
        return run_details

    def is_run_finished(self, run_id):
        """
            method to check if the run is finished
        :param run_id: run_id of the run
        :return: boolean
        """
        status = self.client().get_run(run_id).run.status
        return status in ["Succeeded", "Failed", "Skipped", "Error"]

    def get_run_status(self, run_id):
        """
        method return the status of run
        :param run_id: run_id of the run
        :return: status of the run
        """
        return self.client().get_run(run_id).run.status

    @staticmethod
    def delete_pipeline(pipeline_id):
        """
        method deletes the pipeline
        :param pipeline_id: pipeline id
        :return:
        """
        KubeflowPlugin.client().delete_pipeline(pipeline_id=pipeline_id)

    @staticmethod
    def list_pipeline_versions(pipeline_id):
        """
         method to list the pipeline based on pipeline id
        :param pipeline_id: pipeline id
        :return:
        """
        response = KubeflowPlugin.client().list_pipeline_versions(
            pipeline_id=pipeline_id
        )
        return response

    @staticmethod
    def delete_pipeline_version(version_id):
        """
        method to list the pipeline based on version_id
        :param version_id: pipeline id
        :return:
        """
        KubeflowPlugin.client().delete_pipeline_version(version_id=version_id)

    @staticmethod
    def delete_runs(run_ids):
        """
        delete the pipeline runs
        :param run_ids: list of runs
        :return: successful deletion runs or 404 error
        """
        for run in run_ids:
            KubeflowPlugin.client().runs.delete_run(id=run)



"""
Cogflow module sets up a pipeline for handling datasets and machine learning models
using multiple plugins. It includes functions for creating, registering, evaluating,
and serving models, as well as managing datasets.

Key components include:

Mlflow Plugin: For model tracking, logging, and evaluation.
Kubeflow Plugin: For pipeline management and serving models.
Dataset Plugin: For dataset registration and management.
Model Plugin: For saving model details.
Configurations: Constants for configuration like tracking URIs, database credentials, etc.

Key Functions:

Model Management:

register_model: Register a new model.
log_model: Log a model.
load_model: Load a model.
delete_registered_model: Delete a registered model.
create_registered_model: Create a new registered model.
create_model_version: Create a new version of a registered model.


Run Management

start_run: Start a new.
end_run: End the current.
log_param: Log a parameter to the current run.
log_metric: Log a metric to the current run.

Evaluation and Autologging

evaluate: Evaluate a model.
autolog: Enable automatic logging of parameters, metrics, and models.

Search and Query

search_registered_models: Search for registered models.
search_model_versions: Search for model versions.
get_model_latest_version: Get the latest version of a registered model.
get_artifact_uri: Get the artifact URI of the current or specified run.

Dataset Management

link_model_to_dataset: Link a model to a dataset.
save_dataset_details: Save dataset details.
save_model_details_to_db: Save model details to the database.

Pipeline and Component Management

pipeline: Create a new Kubeflow pipeline.
create_component_from_func: Create a Kubeflow component from a function.
client: Get the Kubeflow client.
load_component_from_url: Load a Kubeflow component from a URL.

Model Serving

serve_model_v1: Serve a model using Kubeflow V1.
serve_model_v2: Serve a model using Kubeflow V2.
get_model_url: Get the URL of a served model.
delete_served_model: Delete a served model.

MinIO Operations

create_minio_client: Create a MinIO client.
query_endpoint_and_download_file: Query an endpoint and download a file from MinIO.
save_to_minio: Save file content to MinIO.
delete_from_minio: Delete an object from MinIO.

Dataset Registration

register_dataset: Register a dataset.
"""

import json
import os
import time
from typing import Union, Any, List, Optional, Dict
import numpy as np
import pandas as pd
import requests
from kfp_server_api import ApiException
from mlflow.models.signature import ModelSignature
from scipy.sparse import csr_matrix, csc_matrix
from .v2 import *
from . import pluginmanager, plugin_config
from .plugin_config import (
    TRACKING_URI,
    TIMER_IN_SEC,
    ML_TOOL,
    ACCESS_KEY_ID,
    SECRET_ACCESS_KEY,
    S3_ENDPOINT_URL,
    ML_DB_USERNAME,
    ML_DB_PASSWORD,
    ML_DB_HOST,
    ML_DB_PORT,
    ML_DB_NAME,
    COGFLOW_DB_USERNAME,
    COGFLOW_DB_PASSWORD,
    COGFLOW_DB_HOST,
    COGFLOW_DB_PORT,
    COGFLOW_DB_NAME,
    MINIO_ENDPOINT_URL,
    MINIO_ACCESS_KEY,
    MINIO_SECRET_ACCESS_KEY,
    API_BASEPATH,
)
from .plugins.dataset_plugin import DatasetMetadata, DatasetPlugin
from .plugins.kubeflowplugin import CogContainer, KubeflowPlugin
from .plugins.mlflowplugin import MlflowPlugin
from .plugins.notebook_plugin import NotebookPlugin
from .pluginmanager import PluginManager
from .util import make_post_request, is_valid_s3_uri

pyfunc = MlflowPlugin().pyfunc
mlflow = MlflowPlugin().mlflow
sklearn = MlflowPlugin().sklearn
cogclient = MlflowPlugin().cogclient
tensorflow = MlflowPlugin().tensorflow
pytorch = MlflowPlugin().pytorch
models = MlflowPlugin().models
lightgbm = MlflowPlugin().lightgbm

add_model_access = CogContainer().add_model_access
kfp = KubeflowPlugin().kfp


def create_minio_client():
    """
    Creates a MinIO client object.

    Returns:
        Minio: The MinIO client object.
    """
    return DatasetPlugin().create_minio_client()


def query_endpoint_and_download_file(url, output_file, bucket_name):
    """
    Queries an endpoint and downloads a file from MinIO.

    Args:
        url (str): The URL to query.
        output_file (str): The output file path.
        bucket_name (str): The MinIO bucket name.

    Returns:
        bool: True if the file was successfully downloaded, False otherwise.
    """
    return DatasetPlugin().query_endpoint_and_download_file(
        url=url, output_file=output_file, bucket_name=bucket_name
    )


def save_to_minio(file_content, output_file, bucket_name):
    """
    Saves file content to MinIO.

    Args:
        file_content (bytes): The content of the file to save.
        output_file (str): The output file path.
        bucket_name (str): The MinIO bucket name.

    Returns:
        bool: True if the file was successfully saved, False otherwise.
    """
    return DatasetPlugin().save_to_minio(
        file_content=file_content, output_file=output_file, bucket_name=bucket_name
    )


def delete_from_minio(object_name, bucket_name):
    """
    Deletes an object from MinIO.

    Args:
        object_name (str): The name of the object to delete.
        bucket_name (str): The MinIO bucket name.

    Returns:
        bool: True if the object was successfully deleted, False otherwise.
    """
    return DatasetPlugin().delete_from_minio(
        object_name=object_name, bucket_name=bucket_name
    )


def register_dataset(details: DatasetMetadata):
    """
    Registers a dataset with the given details.

    Args:
        details (DatasetMetadata): The details of the dataset to register.

    Returns:
        bool: True if the dataset was successfully registered, False otherwise.
    """
    return DatasetPlugin().register_dataset(details=details)


def get_dataset(name: str):
    """
    get a dataset with the given name.
    """
    return DatasetPlugin().get_dataset(name=name)


def delete_registered_model(model_name):
    """
    Deletes a registered model.

    Args:
        model_name (str): The name of the model to delete.

    Returns:
        bool: True if the model was successfully deleted, False otherwise.
    """
    return MlflowPlugin().delete_registered_model(model_name=model_name)


def evaluate(
    data,
    *,
    model_name: str,
    model_uri: str,
    targets,
    model_type: str,
    dataset_path=None,
    feature_names: list = None,
    evaluators=None,
    evaluator_config=None,
    custom_metrics=None,
    custom_artifacts=None,
    validation_thresholds=None,
    baseline_model=None,
    env_manager="local",
):
    """
    Evaluates a model.

    Args:
        model_name: The name of model to evaluate.
        model_uri (str): The URI of the model.
        data: The data to evaluate the model on.
        model_type: The type of the model.
        targets: The targets of the model.
        dataset_path: The path to the dataset.
        feature_names: The names of the features.
        evaluators: The evaluators to use.
        evaluator_config: The configuration for the evaluator.
        custom_metrics: Custom metrics to use.
        custom_artifacts: Custom artifacts to use.
        validation_thresholds: Validation thresholds to use.
        baseline_model: The baseline model to compare against.
        env_manager: The environment manager to use.

    Returns:
        dict: The evaluation results.
    """
    result = MlflowPlugin().evaluate(
        model=model_uri,
        data=data,
        model_type=model_type,
        targets=targets,
        dataset_path=dataset_path,
        feature_names=feature_names,
        evaluators=evaluators,
        evaluator_config=evaluator_config,
        custom_metrics=custom_metrics,
        custom_artifacts=custom_artifacts,
        validation_thresholds=validation_thresholds,
        baseline_model=baseline_model,
        env_manager=env_manager,
    )

    # PluginManager().load_config()
    # # Construct URLs
    # url_metrics = os.getenv(plugin_config.API_BASEPATH) + "/validation/metrics"
    # url_artifacts = os.getenv(plugin_config.API_BASEPATH) + "/validation/artifact"
    #
    # # Attempt to make POST requests, continue regardless of success or failure
    # try:
    #     make_post_request(url_metrics, data=result.metrics, params=model_name)
    # except Exception as exp:
    #     print(f"Failed to post metrics: {exp}")
    #
    # try:
    #     make_post_request(url_artifacts, data=result.artifacts, params=model_name)
    # except Exception as exp:
    #     print(f"Failed to post artifacts: {exp}")

    return result


def search_registered_models(
    filter_string: Optional[str] = None,
    max_results: int = 100,
    order_by: Optional[List[str]] = None,
    page_token: Optional[str] = None,
):
    """
    Searches for registered models.

    This method allows you to search for registered models using optional filtering criteria,
    and retrieve a list of registered models that match the specified criteria.

    Args:
        filter_string (Optional[str]): A string used to filter the registered models. The filter
                string can include conditions on model name, tags and other attributes. For example,
                "name='my_model' AND tags.key='value'". If not provided, all registered
                models are returned.
        max_results (int): The maximum number of results to return. Defaults to 100.
        order_by (Optional[List[str]]): A list of property keys to order the results by.
            For example, ["name ASC", "version DESC"].
        page_token (Optional[str]): A token to specify the page of results to retrieve. This is
            useful for pagination when there are more results than can be returned in a single call.

    Returns:
        List[Dict[str, Any]]: A list of dictionaries, each representing a registered model that
        matches the search criteria. Each dictionary contains details about the registered model,
        such as its name, creation timestamp, last updated timestamp, tags, and description.
    """
    return MlflowPlugin().search_registered_models(
        filter_string=filter_string,
        max_results=max_results,
        order_by=order_by,
        page_token=page_token,
    )


def load_model(model_uri: str, dst_path=None):
    """
    Loads a model from the specified URI.

    Args:
        model_uri (str): The URI of the model to load.
        dst_path (str, optional): The destination path to load the model to.

    Returns:
        Any: The loaded model object.
    """
    return MlflowPlugin().load_model(model_uri=model_uri, dst_path=dst_path)


def register_model(
    model_uri: str,
    model: str,
    await_registration_for: int = 300,
    *,
    tags: Optional[Dict[str, Any]] = None,
):
    """
    Registers the given model with Mlflow.

    This method registers a model with Mlflow using the specified model URI. Optionally,
    tags can be added to the registered model for better organization and metadata tracking.

    Args:
        model_uri (str): The URI of the Mlflow model to register.
        model (str): The name under which to register the model in the Mlflow Model Registry.
        await_registration_for (int, optional): The duration, in seconds, to wait for the model
        version to finish being created and be in the READY status. Defaults to 300 seconds.
        tags (Optional[Dict[str, Any]], optional): A dictionary of key-value pairs to tag the
        registered model with. Tags can be useful for organizing and filtering models in the
         registry.

    Returns:
        ModelVersion: An instance of `ModelVersion` representing the registered model version.
    """
    return MlflowPlugin().register_model(
        model=model,
        model_uri=model_uri,
        await_registration_for=await_registration_for,
        tags=tags,
    )


def autolog():
    """
    Enables automatic logging of parameters, metrics, and models.
    """
    return MlflowPlugin().autolog()


def create_registered_model(
    model: str, tags: Optional[Dict[str, Any]] = None, description: Optional[str] = None
):
    """
    Create a registered model in the Mlflow Model Registry.

    This method creates a new registered model in the Mlflow Model Registry with the given name.
    Optionally, tags and a description can be added to provide additional metadata about the model.

    Args:
        model (str): The name of the registered model.
        tags (Optional[Dict[str, Any]], optional): A dictionary of key-value pairs to tag
        the registered model with. Tags can be useful for organizing and filtering models in the
        registry.
        description (Optional[str], optional): A description of the registered model. This can
        provide additional context about the model's purpose, usage, or any other relevant
        information.

    Returns:
        RegisteredModel: An instance of `RegisteredModel` representing the created registered model.
    """
    return MlflowPlugin().create_registered_model(
        model=model, tags=tags, description=description
    )


def create_model_version(
    model: str,
    source: str,
    run_id: Optional[str] = None,
    tags: Optional[Dict[str, Any]] = None,
    run_link: Optional[str] = None,
    description: Optional[str] = None,
    await_creation_for: int = 300,
):
    """
    Create a model version for a registered model in the Mlflow Model Registry.

    This method registers a new version of an existing registered model with the given
    source path or URI.
    Optionally, additional metadata such as run ID, tags, run link, and description can be provided.
    The `await_creation_for` parameter allows specifying a timeout for waiting for the model version
    creation to complete.

    Args:
        model (str): The name of the registered model.
        source (str): The source path or URI of the model. This is the location where the model
        artifacts are stored.
        run_id (Optional[str], optional): The ID of the run associated with this model version.
            This can be useful for tracking the lineage of the model version.
        tags (Optional[Dict[str, Any]], optional): A dictionary of key-value pairs to tag the
        model version with. Tags can help in organizing and filtering model versions.
        run_link (Optional[str], optional): A URI link to the run. This can provide quick access to
        the run details.
        description (Optional[str], optional): A description of the model version. This can provide
        additional context
            about the changes or improvements in this version.
        await_creation_for (int, optional): The time in seconds to wait for the model version
        creation to complete.
            Defaults to 300 seconds.

    Returns:
        ModelVersion: An instance of `ModelVersion` representing the created model version.
    """
    return MlflowPlugin().create_model_version(
        model=model,
        source=source,
        run_id=run_id,
        tags=tags,
        run_link=run_link,
        description=description,
        await_creation_for=await_creation_for,
    )


def set_tracking_uri(tracking_uri):
    """
    Sets the tracking URI.

    Args:
        tracking_uri (str): The tracking URI to set.
    """
    return MlflowPlugin().set_tracking_uri(tracking_uri=tracking_uri)


def set_experiment(
    experiment_name: Optional[str] = None, experiment_id: Optional[str] = None
):
    """
    Set the active experiment.

    This method sets the specified experiment as the active experiment.
    The active experiment is the one to which subsequent runs will be logged.
    You can specify the experiment by name or by ID.

    Args:
        experiment_name (Optional[str]): The name of the experiment to set as active.
            If `experiment_name` is provided, it takes precedence over `experiment_id`.
        experiment_id (Optional[str]): The ID of the experiment to set as active.
            If `experiment_name` is not provided, `experiment_id` will be used to set
            the active experiment.

    Returns:
        None
    """
    return MlflowPlugin().set_experiment(
        experiment_name=experiment_name, experiment_id=experiment_id
    )


def get_artifact_uri(artifact_path: Optional[str] = None):
    """
    Get the artifact URI of the current or specified  run.

    This method returns the URI of the artifact directory for the current run or for
    the specified artifact path.

    Args:
        artifact_path (Optional[str]): The path of the artifact within the run's artifact directory.
            If not provided, the method returns the URI of the current run's artifact directory.

    Returns:
        str: The URI of the specified artifact path or the current run's artifact directory.
    """
    return MlflowPlugin().get_artifact_uri(artifact_path=artifact_path)


def start_run(
    run_id: Optional[str] = None,
    experiment_id: Optional[str] = None,
    run_name: Optional[str] = None,
    nested: bool = False,
    tags: Optional[Dict[str, Any]] = None,
    description: Optional[str] = None,
):
    """
    Starts a run.

    This method starts a new run or resumes an existing run if a run_id is provided.

    Args:
        run_id (Optional[str]): The ID of the run to resume. If not provided, a new run is started.
        experiment_id (Optional[str]): The ID of the experiment under which to create the run.
        run_name (Optional[str]): The name of the Mlflow run.
        nested (bool): Whether to create the run as a nested run of the parent run.
        tags (Optional[Dict[str, Any]]): A dictionary of tags to set on the run.
        description (Optional[str]): A description for the run.

    Returns:
        The Run object corresponding to the started or resumed run.
    """
    return MlflowPlugin().start_run(
        run_id=run_id,
        experiment_id=experiment_id,
        run_name=run_name,
        nested=nested,
        tags=tags,
        description=description,
    )


def end_run():
    """
    Ends the current run.

    Returns:
        str: The ID of the ended run.
    """
    return MlflowPlugin().end_run()


def log_param(key: str, value: Any):
    """
    Logs a parameter to the current run.

    Args:
        key (str): The key of the parameter.
        value (Any): The value of the parameter.
    """
    return MlflowPlugin().log_param(key=key, value=value)


def log_metric(
    key: str,
    value: float,
    step: Optional[int] = None,
):
    """
    Logs a metric to the current run.

    Args:
        key (str): The key of the metric.
        value (float): The value of the metric.
        step (int, optional): The step at which the metric was logged.
    """
    return MlflowPlugin().log_metric(
        key=key,
        value=value,
        step=step,
    )


def log_model(
    sk_model,
    artifact_path,
    conda_env=None,
    code_paths=None,
    serialization_format="cloudpickle",
    registered_model_name=None,
    signature: ModelSignature = None,
    input_example: Union[
        pd.DataFrame,
        np.ndarray,
        dict,
        list,
        csr_matrix,
        csc_matrix,
        str,
        bytes,
        tuple,
    ] = None,
    await_registration_for=300,
    pip_requirements=None,
    extra_pip_requirements=None,
    pyfunc_predict_fn="predict",
    metadata=None,
):
    """
    Logs a model.

    Args:
        sk_model: The scikit-learn model to log.
        artifact_path (str): The artifact path to log the model to.
        conda_env (str, optional): The conda environment to use.
        code_paths (list, optional): List of paths to include in the model.
        serialization_format (str, optional): The format to use for serialization.
        registered_model_name (str, optional): The name to register the model under.
        signature (ModelSignature, optional): The signature of the model.
        input_example (Union[pd.DataFrame, np.ndarray, dict, list, csr_matrix, csc_matrix, str,
         bytes, tuple], optional): Example input.
        await_registration_for (int, optional): Time to wait for registration.
        pip_requirements (list, optional): List of pip requirements.
        extra_pip_requirements (list, optional): List of extra pip requirements.
        pyfunc_predict_fn (str, optional): The prediction function to use.
        metadata (dict, optional): Metadata for the model.
    """
    result = MlflowPlugin().log_model(
        sk_model=sk_model,
        artifact_path=artifact_path,
        conda_env=conda_env,
        code_paths=code_paths,
        serialization_format=serialization_format,
        registered_model_name=registered_model_name,
        signature=signature,
        input_example=input_example,
        await_registration_for=await_registration_for,
        pip_requirements=pip_requirements,
        extra_pip_requirements=extra_pip_requirements,
        pyfunc_predict_fn=pyfunc_predict_fn,
        metadata=metadata,
    )

    # try:
    #     active_run = mlflow.active_run()
    #     run_id = None
    #     if active_run:
    #         run_id = active_run.info.run_id
    #         print(f"Active run ID: {run_id}")
    #
    #     # Save model details in DB
    #     response = NotebookPlugin().save_model_details_to_db(registered_model_name)
    #     print("response", response)
    #     model_id = response["data"]["id"]
    #     print("model_id", model_id)
    #     if active_run.info.artifact_uri and registered_model_name:
    #         # Construct the model URI
    #         artifact_subpath = "*"
    #         model_uri = f"{active_run.info.artifact_uri}/{registered_model_name}/{artifact_subpath}"
    #         print("model_uri", model_uri)
    #         NotebookPlugin().save_model_uri_to_db(model_id, model_uri)
    # except Exception as exp:
    #     print(f"Failed to log model details to DB: {exp}")

    return result


def log_model_with_dataset(
    sk_model,
    artifact_path,
    dataset: DatasetMetadata,
    conda_env=None,
    code_paths=None,
    serialization_format="cloudpickle",
    registered_model_name=None,
    signature: ModelSignature = None,
    input_example: Union[
        pd.DataFrame,
        np.ndarray,
        dict,
        list,
        csr_matrix,
        csc_matrix,
        str,
        bytes,
        tuple,
    ] = None,
    await_registration_for=300,
    pip_requirements=None,
    extra_pip_requirements=None,
    pyfunc_predict_fn="predict",
    metadata=None,
):
    """
    Logs a model along with its dataset.

    Args:
        sk_model: The scikit-learn model to log.
        artifact_path (str): The artifact path to log the model to.
        dataset (DatasetMetadata): The dataset metadata.
        conda_env (str, optional): The conda environment to use.
        code_paths (list, optional): List of paths to include in the model.
        serialization_format (str, optional): The format to use for serialization.
        registered_model_name (str, optional): The name to register the model under.
        signature (ModelSignature, optional): The signature of the model.
        input_example (Union[pd.DataFrame, np.ndarray, dict, list, csr_matrix, csc_matrix, str,
         bytes, tuple], optional): Example input.
        await_registration_for (int, optional): Time to wait for registration.
        pip_requirements (list, optional): List of pip requirements.
        extra_pip_requirements (list, optional): List of extra pip requirements.
        pyfunc_predict_fn (str, optional): The prediction function to use.
        metadata (dict, optional): Metadata for the model.
    """
    return DatasetPlugin().log_model_with_dataset(
        sk_model=sk_model,
        artifact_path=artifact_path,
        dataset=dataset,
        conda_env=conda_env,
        code_paths=code_paths,
        serialization_format=serialization_format,
        registered_model_name=registered_model_name,
        signature=signature,
        input_example=input_example,
        await_registration_for=await_registration_for,
        pip_requirements=pip_requirements,
        extra_pip_requirements=extra_pip_requirements,
        pyfunc_predict_fn=pyfunc_predict_fn,
        metadata=metadata,
    )


def link_model_to_dataset(dataset_id, model_id):
    """
    Links a model to a dataset.

    Args:
        dataset_id (str): The ID of the dataset.
        model_id (str): The ID of the model.
    """
    return NotebookPlugin().link_model_to_dataset(
        dataset_id=dataset_id, model_id=model_id
    )


def save_model_uri_to_db(model_id, model_uri):
    """
    Save the model URI to the database.

    :param model_id: ID of the model to update.
    :param model_uri: URI of the model to save.
    :return: Response from the database save operation.
    """
    return NotebookPlugin().save_model_uri_to_db(model_id=model_id, model_uri=model_uri)


def save_dataset_details(dataset):
    """
    Saves dataset details.

    Args:
        dataset: The dataset details to save.

    Returns:
        str: Information message confirming the dataset details are saved.
    """
    return DatasetPlugin().save_dataset_details(dataset=dataset)


def save_model_details_to_db(registered_model_name):
    """
    Saves model details to the database.

    Args:
        registered_model_name (str): The name of the registered model.

    Returns:
        str: Information message confirming the model details are saved.
    """
    return NotebookPlugin().save_model_details_to_db(
        registered_model_name=registered_model_name
    )


def get_model_latest_version(registered_model_name):
    """
    Gets the latest version of a registered model.

    Args:
        registered_model_name (str): The name of the registered model.

    Returns:
        str: The latest version of the registered model.
    """
    return NotebookPlugin().get_model_latest_version(
        registered_model_name=registered_model_name
    )


def search_model_versions(
    filter_string: Optional[str] = None,
):
    """
    Searches for model versions in the model registry based on the specified filters.

    Args:
        filter_string (Optional[str], optional): A string specifying the conditions
        that the model versions must meet.
            It is used to filter the model versions. Examples of filter strings
            include "name='my-model'" or "name='my-model' and version='1'".
            If not provided, all model versions are returned.
            Defaults to None.

    Returns:
        List[dict]: A list of dictionaries, each representing a model version that meets
        the filter criteria. Each dictionary contains information about the model version,
        including its name, version number, creation time, run ID, and other metadata.
    """
    return MlflowPlugin().search_model_versions(filter_string=filter_string)


def pipeline(name=None, description=None):
    """
    Creates a new Kubeflow pipeline.

    Args:
        name (str, optional): The name of the pipeline.
        description (str, optional): The description of the pipeline.

    Returns:
        str: Information message confirming the pipeline creation.
    """
    return KubeflowPlugin().pipeline(name=name, description=description)


def create_component_from_func(
    func,
    output_component_file=None,
    base_image=None,
    packages_to_install=None,
):
    """
    Creates a Kubeflow component from a function.

    Args:
        func: The function to create the component from.
        output_component_file (str, optional): The output file for the component.
        base_image (str, optional): The base image to use.
        packages_to_install (list, optional): List of packages to install.

    Returns:
        str: Information message confirming the component creation.
    """
    return KubeflowPlugin().create_component_from_func(
        func=func,
        output_component_file=output_component_file,
        base_image=base_image,
        packages_to_install=packages_to_install,
    )


def client():
    """
    Gets the Kubeflow client.

    Returns:
        KubeflowClient: The Kubeflow client object.
    """
    return KubeflowPlugin().client()


def input_path(label: str):
    """
    Gets the input path for a Kubeflow component.

    Args:
        label (str): The label for the input path.

    Returns:
        str: The input path.
    """
    return KubeflowPlugin().input_path(label=label)


def output_path(label: str):
    """
    Gets the output path for a Kubeflow component.

    Args:
        label (str): The label for the output path.

    Returns:
        str: The output path.
    """
    return KubeflowPlugin().output_path(label=label)


def serve_model_v2(model_uri: str, name: str = None):
    """
    Serves a model using Kubeflow V2.

    Args:
        model_uri (str): The URI of the model to serve.
        name (str, optional): The name of the model to serve.

    Returns:
        str: Information message confirming the model serving.
    """
    return KubeflowPlugin().serve_model_v2(model_uri=model_uri, name=name)


def serve_model_v1(model_uri: str, name: str = None):
    """
    Serves a model using Kubeflow V1.

    Args:
        model_uri (str): The URI of the model to serve.
        name (str, optional): The name of the model to serve.

    Returns:
        str: Information message confirming the model serving.
    """
    return KubeflowPlugin().serve_model_v1(model_uri=model_uri, name=name)


def get_model_url(model_name: str):
    """
    Gets the URL of a served model.

    Args:
        model_name (str): The name of the served model.

    Returns:
        str: The URL of the served model.
    """
    return KubeflowPlugin().get_model_url(model_name=model_name)


def delete_served_model(model_name: str):
    """
    Deletes a served model.

    Args:
        model_name (str): The name of the model to delete.

    Returns:
        str: Information message confirming the deletion of the served model.
    """
    return KubeflowPlugin().delete_served_model(model_name=model_name)


def load_component(file_path=None, url=None, text=None):
    """Loads component from text, file or URL and creates a task factory
    function.

    Only one argument should be specified.

    Args:
        file_path: Path of local file containing the component definition.
        url: The URL of the component file data.
        text: A string containing the component file data.

    Returns:
        A factory function with a strongly-typed signature.
        Once called with the required arguments, the factory constructs a pipeline task instance (ContainerOp).
    """
    non_null_args_count = len(
        [name for name, value in locals().items() if value is not None]
    )
    if non_null_args_count != 1:
        raise ValueError("Need to specify exactly one source")
    if file_path:
        return KubeflowPlugin().load_component_from_file(file_path=file_path)
    if url:
        return KubeflowPlugin().load_component_from_url(url=url)
    if text:
        return KubeflowPlugin().load_component_from_text(text=text)
    raise ValueError("Need to specify a source")


def delete_pipeline(pipeline_id):
    """
    method deletes the pipeline
    :param pipeline_id: pipeline id
    :return:
    """
    # list the runs based on pipeline_id
    run_info = NotebookPlugin.list_runs_by_pipeline_id(pipeline_id)
    run_ids = [run["uuid"] for run in run_info["data"]]

    # delete the runs from kfp and db based on pipeline_id
    try:
        KubeflowPlugin().delete_runs(run_ids)
        NotebookPlugin.delete_run_details_from_db(pipeline_id)
    except ApiException as exp:
        print(f"Failed to delete run for the pipeline id {pipeline_id}: {exp}")

    # list the pipeline versions and delete from kfp
    pipeline_version_response = KubeflowPlugin().list_pipeline_versions(
        pipeline_id=pipeline_id
    )
    if pipeline_version_response.versions:
        pipeline_version_details = pipeline_version_response.versions

        pipeline_version_ids = [version.id for version in pipeline_version_details]
        print("Pipeline Version IDs to delete:", pipeline_version_ids)

        # Delete each pipeline version
        for version_id in pipeline_version_ids:
            try:
                KubeflowPlugin().delete_pipeline_version(version_id=version_id)
                print(f"Deleted pipeline version: {version_id}")
            except ApiException as exp:
                print(f"Failed to delete pipeline version {version_id}: {exp}")
    else:
        print(
            f"No pipeline versions found for the specified pipeline ID {pipeline_id}."
        )

    # Delete the pipeline itself
    try:
        KubeflowPlugin().delete_pipeline(pipeline_id=pipeline_id)
        print(f"Deleted pipeline: {pipeline_id}")
    except ApiException as exp:
        print(f"Failed to delete pipeline {pipeline_id}: {exp}")

    NotebookPlugin.delete_pipeline_details_from_db(pipeline_id)


def cogcomponent(
    output_component_file=None, base_image=None, packages_to_install=None
):
    """
    Decorator to create a Kubeflow component from a function.

    Args:
        output_component_file (str, optional): The output file for the component.
        base_image (str, optional): The base image to use.
        packages_to_install (list, optional): List of packages to install.

    Returns:
        function: The decorator function that wraps the original function.
    """

    def decorator(func):
        component = KubeflowPlugin().create_component_from_func(
            func=func,
            output_component_file=output_component_file,
            base_image=base_image,
            packages_to_install=packages_to_install
        )

        return component

    return decorator


def create_run_from_pipeline_func(
    pipeline_func: ...,
    arguments: Optional[Dict[str, Any]] = None,
    run_name: Optional[str] = None,
    experiment_name: Optional[str] = None,
    namespace: Optional[str] = None,
    pipeline_root: Optional[str] = None,
    enable_caching: Optional[bool] = None,
    service_account: Optional[str] = None,
):
    """
        method to create a run from pipeline function
    :param pipeline_func:
    :param arguments:
    :param run_name:
    :param experiment_name:
    :param namespace:
    :param pipeline_root:
    :param enable_caching:
    :param service_account:
    :return:
    """
    run_details = KubeflowPlugin().create_run_from_pipeline_func(
        pipeline_func=pipeline_func,
        arguments=arguments,
        run_name=run_name,
        experiment_name=experiment_name,
        namespace=namespace,
        pipeline_root=pipeline_root,
        enable_caching=enable_caching,
        service_account=service_account,
    )
    # Poll the run status
    while not KubeflowPlugin().is_run_finished(run_details.run_id):
        status = KubeflowPlugin().get_run_status(run_details.run_id)
        print(f"Run {run_details.run_id} status: {status}")
        time.sleep(plugin_config.TIMER_IN_SEC)

    details = get_pipeline_and_experiment_details(run_details.run_id)
    print("details of upload pipeline", details)
    NotebookPlugin().save_pipeline_details_to_db(details)
    return run_details


def get_pipeline_and_experiment_details(run_id):
    """
        method to return pipeline,run_details,task_details,experiments details based on run_id
    :param run_id: run_id of the run
    :return: dictionary with all the details of pipeline,run_details,task_details,experiments
    """
    try:
        # Get the run details using the run_id
        run_detail = KubeflowPlugin().client().get_run(run_id)
        # Extract run details
        run = run_detail.run
        pipeline_id = run.pipeline_spec.pipeline_id
        experiment_id = run.resource_references[0].key.id
        run_details = {
            "uuid": run.id,
            "display_name": run.name,
            "name": run.name,
            "description": run.description,
            "experiment_uuid": experiment_id,
            "pipeline_uuid": pipeline_id,
            "createdAt_in_sec": run.created_at,
            "scheduledAt_in_sec": run.scheduled_at,
            "finishedAt_in_sec": run.finished_at,
            "state": run.status,
        }

        # Get experiment details using the experiment_id
        experiment = (
            KubeflowPlugin().client().get_experiment(experiment_id=experiment_id)
        )

        experiment_details = {
            "uuid": experiment.id,
            "name": experiment.name,
            "description": experiment.description,
            "createdAt_in_sec": experiment.created_at,
        }

        # Get pipeline details using the pipeline_id
        pipeline_info = KubeflowPlugin().client().get_pipeline(pipeline_id=pipeline_id)

        pipeline_details = {
            "uuid": pipeline_info.id,
            "createdAt_in_sec": pipeline_info.created_at,
            "name": pipeline_info.name,
            "description": pipeline_info.description,
            "parameters": pipeline_info.parameters,
            "experiment_uuid": experiment.id,
            "pipeline_spec": run.pipeline_spec.workflow_manifest,
            "status": run.status,
        }

        workflow_manifest = run_detail.pipeline_runtime.workflow_manifest
        workflow = json.loads(workflow_manifest)

        # Extract the task details
        tasks = workflow["status"]["nodes"]

        task_details = []
        for task_id, task_info in tasks.items():
            task_detail = {
                "uuid": task_id,
                "name": task_info.get("displayName"),
                "state": task_info.get("phase"),
                "runuuid": run.id,
                "startedtimestamp": task_info.get("startedAt"),
                "finishedtimestamp": task_info.get("finishedAt"),
                "createdtimestamp": task_info.get("createdAt"),
            }
            task_details.append(task_detail)

        steps = workflow["status"]["nodes"]
        model_uris = []

        for step_name, step_info in steps.items():
            print(f"step={step_name}")
            if step_info["type"] == "Pod":
                outputs = step_info.get("outputs", {}).get("parameters", [])
                for output in outputs:
                    print(f"Artifact: {output['name']}")
                    print(f"URI: {output['value']}")
                    if is_valid_s3_uri(output["value"]):
                        model_uris.append(output["value"])
                    else:
                        print("Not valid model-uri")
        model_uris = list(set(model_uris))

        model_ids = []
        for model_uri in model_uris:
            PluginManager().load_config()
            # Define the URL
            url = os.getenv(plugin_config.API_BASEPATH) + "/models/uri"
            data = {"uri": model_uri}
            json_data = json.dumps(data)
            headers = {"Content-Type": "application/json"}
            # Make the GET request
            response = requests.get(url, data=json_data, headers=headers)

            # Check if the request was successful
            if response.status_code == 200:
                # Print the response content
                # print('Response Content:')
                model_ids.append(response.json()["data"])
            else:
                print(f"Failed to retrieve data: {response.status_code}")

        return {
            "run_details": run_details,
            "experiment_details": experiment_details,
            "pipeline_details": pipeline_details,
            "task_details": task_details,
            "model_ids": model_ids,
        }
    except Exception as e:
        print(e)


__all__ = [
    # Methods from MlflowPlugin class
    "pyfunc",
    "mlflow",
    "sklearn",
    "cogclient",
    "tensorflow",
    "pytorch",
    "models",
    "lightgbm",
    # Method from CogContainer class
    "add_model_access",
    "kfp",
    "v2",
]



"""
    Utility functions
"""

import re
from datetime import datetime

import requests


def make_post_request(url, data=None, params=None, files=None):
    """
        utility function to make POST requests
    :param url: url of the API endpoint
    :param data: JSON payload
    :param params: request params
    :param files : file
    :return: response for the POST request in json format
    """
    try:
        if data:
            response = requests.post(url, json=data, params=params)
        elif files:
            with open(files, "rb") as file_data:
                file = {"file": file_data}
                response = requests.post(url, params=params, files=file)
                file_data.close()
        else:
            response = requests.post(url, params=params)

        if response.status_code == 201:
            print("POST request successful")
            return response.json()
        # if not the success response
        print(f"POST request failed with status code {response.status_code}")
        raise Exception("request failed")
    except requests.exceptions.RequestException as exp:
        print(f"Error making POST request: {exp}")
        raise Exception(f"Error making POST request: {exp}")


def custom_serializer(obj):
    """
        method to serializer obj to datetime isoformat
    :param obj:
    :return:
    """
    if isinstance(obj, datetime):
        return obj.isoformat()
    raise TypeError(f"Type {type(obj)} not serializable")


def is_valid_s3_uri(uri):
    """
        method to check if the url is valid S3 url
    :param uri: url to check
    :return:
    """
    # Regular expression for S3 URI
    s3_uri_regex = re.compile(r"^s3://([a-z0-9.-]+)/(.*)$")

    # Check against the regex pattern
    match = s3_uri_regex.match(uri)

    if match:
        bucket_name = match.group(1)
        object_key = match.group(2)

        # Additional checks for bucket name and object key can be added here
        if bucket_name and object_key:
            return True

    return False


def make_delete_request(url, path_params=None, query_params=None):
    """
     utility function to make DELETE requests
    :url: url of the API endpoint
    :path_params: path params
    :query_params: query params
    :return: response for the POST request in json format
    """
    try:
        if query_params:
            response = requests.delete(url, params=query_params)
        else:
            # Make the DELETE request with query parameters
            response = requests.delete(url + "/" + path_params)
        if response.status_code == 200:
            print("DELETE request successful")
            return response.json()
        # if not the success response
        print(f"DELETE request failed with status code {response.status_code}")
        raise Exception("request failed")
    except requests.exceptions.RequestException as exp:
        print(f"Error making POST request: {exp}")
        raise Exception(f"Error making POST request: {exp}")


def make_get_request(url, path_params=None, query_params=None):
    """
     utility function to make GET requests
    :url: url of the API endpoint
    :path_params: path params
    :query_params: query params
    :return: response for the GET request in json format
    """

    try:
        if query_params:
            response = requests.get(url, params=query_params)
        else:
            # Make the GET request with query parameters
            response = requests.get(url + "/" + path_params)
        if response.status_code == 200:
            print("GET request successful")
            return response.json()
        # if not the success response
        print(f"GET request failed with status code {response.status_code}")
        raise Exception("request failed")
    except requests.exceptions.RequestException as exp:
        print(f"Error making GET request: {exp}")
        raise Exception(f"Error making GET request: {exp}")



import json
import unittest
from contextlib import redirect_stdout
from io import StringIO
from unittest.mock import patch, MagicMock

import cogflow.cogflow.plugins.notebook_plugin
from cogflow.cogflow import (
    link_model_to_dataset,
    save_dataset_details,
    save_model_details_to_db,
)
from ..cogflow.plugins.dataset_plugin import DatasetMetadata, DatasetPlugin
from ..cogflow.plugins.notebook_plugin import NotebookPlugin


class TestNotebookPlugin(unittest.TestCase):
    @patch(
        "cogflow.cogflow.plugins.notebook_plugin.NotebookPlugin.get_model_latest_version"
    )
    @patch("os.getenv")
    def test_save_model_details_to_db(self, mock_env, mock_model_version):
        with patch("requests.post") as mock_requests_post:
            mock_env.side_effect = lambda x: {
                "MLFLOW_S3_ENDPOINT_URL": "localhost:9000",
                "AWS_ACCESS_KEY_ID": "minio",
                "AWS_SECRET_ACCESS_KEY": "minio123",
                "API_BASEPATH": "http://randomn",
            }[x]
            mock_model_version.return_value = 1

            mock_response = {
                "data": {
                    "id": 101,
                    "last_modified_time": "2024-05-16T12:33:08.890033",
                    "last_modified_user_id": 0,
                    "name": "testmodel",
                    "register_date": "2024-05-16T12:33:08.890007",
                    "register_user_id": 0,
                    "type": "sklearn",
                    "version": "1",
                },
                "errors": "None",
                "message": "Created new model.",
                "success": "True",
            }
            mock_requests_post.return_value.status_code = 201
            mock_requests_post.return_value.json.return_value = mock_response
            result = save_model_details_to_db("testmodel")
            self.assertEqual(result["data"]["id"], 101)

    @patch("os.getenv")
    def test_save_dataset_details(self, mock_env):
        with patch("requests.post") as mock_requests_post:
            mock_env.side_effect = lambda x: {
                "MLFLOW_S3_ENDPOINT_URL": "localhost:9000",
                "AWS_ACCESS_KEY_ID": "minio",
                "AWS_SECRET_ACCESS_KEY": "minio123",
                "API_BASEPATH": "http://randomn",
            }[x]

            mock_response = {
                "data": {
                    "dataset_id": 8,
                    "file_name": "breastcancerwisconsindiagnostic.zip",
                    "file_path": "mlflow",
                    "register_date": "2024-05-16T13:03:05.442386",
                    "user_id": 0,
                },
                "errors": "None",
                "message": "File uploaded successfully.",
                "success": "True",
            }
            mock_requests_post.return_value.status_code = 201
            mock_requests_post.return_value.json.return_value = mock_response
            # Dataset details
            source = (
                "https://archive.ics.uci.edu/static/public/17"
                "/breast+cancer+wisconsin+diagnostic.zip"
            )
            format = "zip"
            name = "breast+cancer+wisconsin+diagnostic.zip"
            description = "Breast cancer wisconsin diagnotic dataset"

            dm = DatasetMetadata(name, description, source, format)
            result = save_dataset_details(dataset=dm)
            self.assertEqual(result, mock_response["data"]["dataset_id"])

    @patch("os.getenv")
    def test_link_model_to_dataset(self, mock_env):
        with patch("requests.post") as mock_requests_post:
            mock_env.side_effect = lambda x: {
                "MLFLOW_S3_ENDPOINT_URL": "localhost:9000",
                "AWS_ACCESS_KEY_ID": "minio",
                "AWS_SECRET_ACCESS_KEY": "minio123",
                "API_BASEPATH": "http://randomn",
            }[x]

            mock_response = {
                "data": {
                    "dataset_id": 2,
                    "linked_time": "2024-05-16 15:23:24",
                    "model_id": 1,
                    "user_id": 0,
                },
                "errors": "None",
                "message": "Dataset linked with model successfully",
                "success": "True",
            }
            mock_requests_post.return_value.status_code = 201
            mock_requests_post.return_value.json.return_value = mock_response
            f = StringIO()
            with redirect_stdout(f):
                link_model_to_dataset(2, 1)
            out = f.getvalue().strip()
            assert out == "POST request successful"

    @patch("os.getenv")
    def test_delete_pipeline_details_from_db(self, mock_env):
        with patch("requests.delete") as mock_requests_delete:
            mock_env.side_effect = lambda x: {
                "MLFLOW_S3_ENDPOINT_URL": "localhost:9000",
                "AWS_ACCESS_KEY_ID": "minio",
                "AWS_SECRET_ACCESS_KEY": "minio123",
                "API_BASEPATH": "http://randomn",
            }[x]

            mock_response = {
                "errors": "None",
                "message": "Pipeline Details Deleted successfully",
                "success": "True",
            }
            mock_requests_delete.return_value.status_code = 200
            mock_requests_delete.return_value.json.return_value = mock_response
            f = StringIO()
            with redirect_stdout(f):
                NotebookPlugin().delete_pipeline_details_from_db("2")
            out = f.getvalue().strip()
            assert out == "DELETE request successful"

    @patch("os.getenv")
    def test_delete_run_details_from_db(self, mock_env):
        with patch("requests.delete") as mock_requests_delete:
            mock_env.side_effect = lambda x: {
                "MLFLOW_S3_ENDPOINT_URL": "localhost:9000",
                "AWS_ACCESS_KEY_ID": "minio",
                "AWS_SECRET_ACCESS_KEY": "minio123",
                "API_BASEPATH": "http://randomn",
            }[x]

            mock_response = {
                "errors": "None",
                "message": "Runs deleted successfully",
                "success": "True",
            }
            mock_requests_delete.return_value.status_code = 200
            mock_requests_delete.return_value.json.return_value = mock_response
            f = StringIO()
            with redirect_stdout(f):
                NotebookPlugin().delete_run_details_from_db("2")
            out = f.getvalue().strip()
            assert out == "DELETE request successful"

    @patch("os.getenv")
    def test_list_runs_by_pipeline_id(self, mock_env):
        with patch("requests.get") as mock_requests_get:
            mock_env.side_effect = lambda x: {
                "MLFLOW_S3_ENDPOINT_URL": "localhost:9000",
                "AWS_ACCESS_KEY_ID": "minio",
                "AWS_SECRET_ACCESS_KEY": "minio123",
                "API_BASEPATH": "http://randomn",
            }[x]

            # Define mock response
            mock_response = {
                "runs": [
                    {"run_id": "0d3ffa58-7d15-4456-a1f6-2c1355f95d22"},
                    {"run_id": "0d3ffa58-7d15-4456-a1f6-2c1355f95d23"},
                ]
            }

            mock_requests_get.return_value.status_code = 200
            mock_requests_get.return_value.json.return_value = mock_response
            f = StringIO()
            with redirect_stdout(f):
                NotebookPlugin().list_runs_by_pipeline_id("2")
            out = f.getvalue().strip()
            assert out == "GET request successful"

    @patch("cogflow.cogflow.plugins.notebook_plugin.NotebookPlugin.load_pkl")
    @patch(
        "cogflow.cogflow.plugins.notebook_plugin.NotebookPlugin.get_model_latest_version"
    )
    @patch("mlflow.sklearn.log_model")
    def test_log_model_by_model_file(
        self, mock_log_model, mock_model_version, mock_load_pkl
    ):
        sk_model = MagicMock()
        mock_load_pkl.return_value = sk_model
        mock_model_version.return_value = 1
        model_file_path = "var/data/model.pkl"
        model_name = "test_model"
        result = NotebookPlugin().log_model_by_model_file(model_file_path, model_name)
        self.assertEqual(result["version"], 1)
        self.assertIn("artifact_uri", result)
        mock_log_model.assert_called_once()

    def test_deploy_model(self):
        with patch(
            "cogflow.cogflow.plugins.mlflowplugin.MlflowPlugin.get_model_uri"
        ) as mock_model_uri:
            with patch(
                "cogflow.cogflow.plugins.kubeflowplugin.KubeflowPlugin.serve_model_v1"
            ) as mock_serve:
                result = NotebookPlugin().deploy_model("Flearning", "1", "fl-svc")
                self.assertEqual(result["status"], True)
                self.assertEqual(
                    result["msg"], "Model Flearning deployed with service fl-svc"
                )

    def test_deploy_model_for_model_not_found_exception(self):
        with patch(
            "cogflow.cogflow.plugins.mlflowplugin.MlflowPlugin.get_model_uri"
        ) as mock_model_uri:
            mock_model_uri.side_effect = Exception()
            with self.assertRaises(Exception):
                NotebookPlugin().deploy_model("Flearning", "1", "fl-svc")

    @patch.object(
        cogflow.cogflow.plugins.notebook_plugin.NotebookPlugin, "run_kubectl_command"
    )
    def test_get_pods_in_namespace(self, mock_run_kubectl_command):
        mock_output = {
            "items": [{"metadata": {"name": "pod1"}}, {"metadata": {"name": "pod2"}}]
        }
        mock_run_kubectl_command.return_value = json.dumps(mock_output)

        pods = NotebookPlugin.get_pods_in_namespace("test-namespace")

        self.assertEqual(len(pods), 2)
        self.assertEqual(pods[0]["metadata"]["name"], "pod1")
        self.assertEqual(pods[1]["metadata"]["name"], "pod2")

    @patch.object(
        cogflow.cogflow.plugins.notebook_plugin.NotebookPlugin, "run_kubectl_command"
    )
    def test_get_pod_logs(self, mock_run_kubectl_command):
        mock_logs = "Mock pod logs"
        mock_run_kubectl_command.return_value = mock_logs

        logs = NotebookPlugin.get_pod_logs("test-namespace", "test-pod")

        self.assertEqual(logs, mock_logs)

    @patch.object(
        cogflow.cogflow.plugins.notebook_plugin.NotebookPlugin, "run_kubectl_command"
    )
    def test_get_pod_prefix(self, mock_run_kubectl_command):
        mock_output = {
            "status": {
                "components": {"predictor": {"latestReadyRevision": "revision-123"}}
            }
        }
        mock_run_kubectl_command.return_value = json.dumps(mock_output)

        prefix = NotebookPlugin.get_pod_prefix("test-inference-service")

        self.assertEqual(prefix, "revision-123")

    @patch.object(
        cogflow.cogflow.plugins.notebook_plugin.NotebookPlugin, "get_pods_in_namespace"
    )
    @patch.object(
        cogflow.cogflow.plugins.notebook_plugin.NotebookPlugin, "get_pod_prefix"
    )
    @patch.object(
        cogflow.cogflow.plugins.notebook_plugin.NotebookPlugin, "get_pod_logs"
    )
    @patch.object(json, "dump")
    def test_get_logs_for_inference_service(
        self,
        mock_json_dump,
        mock_get_pod_logs,
        mock_get_pod_prefix,
        mock_get_pods_in_namespace,
    ):
        mock_get_pod_prefix.return_value = "revision-123"
        mock_get_pods_in_namespace.return_value = [
            {"metadata": {"name": "revision-123-pod1"}},
            {"metadata": {"name": "revision-123-pod2"}},
        ]
        mock_get_pod_logs.side_effect = (
            lambda namespace, pod_name: f"Logs for {pod_name}"
        )

        logs_output = NotebookPlugin.get_logs_for_inference_service(
            "test-namespace", "test-inference-service"
        )

        self.assertIn("revision-123-pod1", logs_output)
        self.assertIn("revision-123-pod2", logs_output)
        self.assertEqual(logs_output["revision-123-pod1"], "Logs for revision-123-pod1")
        self.assertEqual(logs_output["revision-123-pod2"], "Logs for revision-123-pod2")


if __name__ == "__main__":
    unittest.main()



"""
    This file contains unittest for PluginManager methods
"""

import os
import unittest
import configparser
from unittest.mock import patch, mock_open, MagicMock
from ..cogflow.pluginmanager import PluginManager


class TestPluginManager(unittest.TestCase):
    """
    Test class for PluginManger
    """

    def setUp(self):
        # Set up the file paths and sections to use in the test cases
        self.config_file_path = os.path.join(
            os.path.dirname(__file__), "cogflow_config.ini"
        )
        # self.config_file_path = "config.ini"
        self.section = "mlflow_plugin"
        self.key = "activation_key"

        # Initialize the instance of the class containing the methods to test
        self.instance = PluginManager()

    def test_get_config_value_activation_status(self):
        # Prepare a valid configuration file content with the key
        config_content = """
        [mlflow_plugin]
        activation_key = True
        """

        # Mock open function to simulate reading the file
        with patch("builtins.open", mock_open(read_data=self.config_file_path)):
            # Mock configparser to control its behavior
            with patch("configparser.ConfigParser") as mock_config:
                # Configure the mock object to return desired values
                mock_config_instance = mock_config.return_value
                mock_config_instance.read.return_value = None
                mock_config_instance.getboolean.return_value = True
                print("**************", self.config_file_path)

                # Call the method

                with self.assertRaises(Exception) as context:
                    self.instance.get_config_value(
                        self.config_file_path, "mlflow_plugin"
                    )

                # Check that the expected error message is raised
                expected_message = (
                    "Section 'mlflow_plugin' not found in the configuration file. "
                    "Please correct section name in configuration file."
                )
                self.assertIn(expected_message, str(context.exception))

    def test_get_config_value_nonexistent_key(self):
        # Prepare a valid configuration file content without the key
        config_content = """
        [mlflow_plugin]
        other_key = some_value
        """

        # Mock open function to simulate reading the file
        with patch("builtins.open", mock_open(read_data=config_content)):
            # Mock configparser to control its behavior
            with patch("configparser.ConfigParser") as mock_config:
                # Configure the mock object to return desired values
                mock_config_instance = mock_config.return_value
                mock_config_instance.read.return_value = None
                mock_config_instance.get.side_effect = configparser.NoOptionError(
                    "key", self.section
                )

                # Call the method and expect a KeyError
                with self.assertRaises(Exception) as cm:
                    self.instance.get_config_value(
                        self.config_file_path, self.section, self.key
                    )
                self.assertNotIn("Key 'activation_key' not found", str(cm.exception))

    def test_verify_activation_plugin_active(self):
        # Prepare a valid configuration file content with the key set to True
        config_content = """
        [mlflow_plugin]
        activation_key = true
        """

        # Mock open function to simulate reading the file
        with patch("builtins.open", mock_open(read_data=config_content)):
            # Mock configparser to control its behavior
            with patch("configparser.ConfigParser") as mock_config:
                # Configure the mock object to return desired values
                mock_config_instance = mock_config.return_value
                mock_config_instance.read.return_value = None
                mock_config_instance.getboolean.return_value = True

                # Call the method
                with self.assertRaises(Exception):
                    self.instance.verify_activation(self.section)
                # No exception expected if plugin is active



"""
Plugin Manager Module

This module provides a PluginManager class responsible for managing plugins such as MlflowPlugin,
KubeflowPlugin, and DatasetPlugin.
It also includes functions to activate, deactivate, and check the status of plugins.

Attributes:
"""

import os
import configparser
import pandas as pd
from sqlalchemy import create_engine, Column, FLOAT, BIGINT, String
from sqlalchemy.orm import declarative_base
from . import plugin_config

Base = declarative_base()


class ModelTraining(Base):
    """
    SQLAlchemy ORM class representing the 'Model_training' table in the PostgreSQL database.

    This table stores information related to model training, including parameters, metrics,
    and other metadata associated with training runs.

    Attributes:
        param_key (str): Key for the parameter.
        param_value (str): Value of the parameter.
        model_name (str): Name of the model.
        model_version (float): Version of the model.
        creation_time (int): Creation time of the model, represented as a Unix timestamp.
        metric_key (str): Key for the metric.
        metric_value (float): Value of the metric.
        run_name (str): Name of the training run.
        run_uuid (str): UUID of the training run, serves as the primary key.
        user_id (str): Identifier for the user who initiated the training run.
    """

    __tablename__ = "Model_training"
    param_key = Column(String)
    param_value = Column(String)
    model_name = Column(String)
    model_version = Column(FLOAT)
    creation_time = Column(BIGINT)
    metric_key = Column(String)
    metric_value = Column(FLOAT)
    run_name = Column(String)
    run_uuid = Column(String, primary_key=True)
    user_id = Column(String)


class PluginManager:
    """
    Class responsible for managing plugins.

    Attributes:
    """

    def __init__(self):
        """
        Initializes the PluginManager with plugin classes.
        """
        self.config_file_path = os.path.join(
            os.path.dirname(__file__), "cogflow_config.ini"
        )

    @staticmethod
    def connect_to_cogflow_db(dialect="postgresql+psycopg2"):
        """
        Create and return an SQLAlchemy engine for a any database.

        Args:
            dialect (str): The connection string for the database. By default postgres DB

        Returns:
            sqlalchemy.engine.Engine: The SQLAlchemy engine.
        """
        postgres_connection_str = (
            f"{dialect}://{plugin_config.COGFLOW_DB_USERNAME}:"
            f"{plugin_config.COGFLOW_DB_PASSWORD}@{plugin_config.COGFLOW_DB_HOST}:"
            f"{plugin_config.COGFLOW_DB_PORT}/{plugin_config.COGFLOW_DB_NAME}"
        )
        engine = create_engine(postgres_connection_str)
        engine.connect()
        return engine

    def insert_data_to_table(self, if_exists: str = "append"):
        """
        Insert data from a pandas DataFrame into a PostgreSQL table.

        Args:
            df (pd.DataFrame): The DataFrame containing data to insert.
            engine: The SQLAlchemy engine to use for the connection.
            table_name (str): The name of the PostgreSQL table to insert data into.
            if_exists (str): Behavior if the table already exists. 'append' to add data,
                             'replace' to drop and recreate the table. Defaults to 'append'.

        Returns:
            None
        """
        data_frame = self.get_params_and_metrics_from_mlflow_db()
        engine = self.connect_to_cogflow_db()
        table_name = ModelTraining.__tablename__
        data_frame.to_sql(table_name, engine, if_exists=if_exists, index=False)

    @staticmethod
    def connect_to_mlflow_db(dialect="mysql+pymysql"):
        """
        Establishes a connection to the MLflow database.

        This method creates an SQLAlchemy engine using the provided database dialect and connection
        parameters from the configuration. It then establishes a connection to the MLflow database
        and returns the engine.

        Args:
            dialect (str): The database dialect to use for the connection (e.g., "mysql+pymysql").
                           Defaults to "mysql+pymysql".

        Returns:
            sqlalchemy.engine.Engine: The SQLAlchemy engine connected to the MLflow database.
        """

        connection_str = (
            f"{dialect}://{plugin_config.ML_DB_USERNAME}:{plugin_config.ML_DB_PASSWORD}"
            f"@{plugin_config.ML_DB_HOST}:{plugin_config.ML_DB_PORT}/{plugin_config.ML_DB_NAME}"
        )
        engine = create_engine(connection_str)
        engine.connect()
        return engine

    def get_params_and_metrics_from_mlflow_db(self):
        """
        Retrieves parameters and metrics from the MLflow database.

        This method queries the MLflow database to retrieve parameters, metrics, and associated
        information such as model name, version, creation time, and other metadata for each run.
        The method executes a SQL query to join multiple tables and returns the results as a pandas
        DataFrame.

        Returns:
            pd.DataFrame: A pandas DataFrame containing the parameters and metrics data from the
                          MLflow database. Each row represents a run with columns for the
                          parameter key,
                          parameter value, model name, model version, creation time, metric key,
                          metric value, run name, run UUID, and user ID.
        """

        metric_query = """
        SELECT p.key AS param_key, p.value AS param_value,
               mv.name AS model_name, mv.version AS model_version, mv.creation_time,
               m.key AS metric_key, m.value AS metric_value,
               r.name AS run_name, r.run_uuid AS run_uuid, r.user_id
        FROM params p
        JOIN model_versions mv ON p.run_uuid = mv.run_id
        JOIN metrics m ON p.run_uuid = m.run_uuid
        JOIN runs r ON p.run_uuid = r.run_uuid;
        """
        engine = self.connect_to_mlflow_db()

        data_frame = pd.read_sql_query(metric_query, engine)
        return data_frame

    @staticmethod
    def get_config_value(config_file_path, section, key="activation_key"):
        """
        Reads the activation status of a plugin from an INI configuration file.

        Args:
            config_file_path (str): The path to the INI configuration file.
            section (str): The section in the INI file that contains the plugin settings.
            key (str, optional): The key used to store activation status in the section.
            Default is 'activation_key'.

        Returns:
            bool: True if the plugin is activated, False otherwise.
            str or None: The value of the key if no exceptions occur and the plugin is activated.
        """
        if not config_file_path:
            raise FileNotFoundError("Configuration file path not provided.")

        # Create a ConfigParser instance
        config = configparser.ConfigParser()

        try:
            # Read the INI configuration file
            config.read(config_file_path)

            # Check if the configuration file is empty or improperly formatted
            if not config.sections():
                raise Exception(
                    "Configuration file is empty or not properly formatted."
                )

            # Check if the specified section exists in the configuration file
            if section not in config:
                raise KeyError(
                    f"Section '{section}' not found in the configuration file. "
                    f"Please correct section name in configuration file."
                )

            value = config.get(section, key, fallback=None)

            # Check if the key value can be converted to boolean
            if value is not None:
                try:
                    # Try to convert the value to boolean
                    activation_status = config.getboolean(section, key, fallback=False)
                    return activation_status
                except ValueError:
                    # If conversion to boolean fails, return the value as a string
                    return value
            else:
                raise KeyError(
                    f"Key '{key}' not found in section '{section}' of the configuration file."
                    f" Please correct key name in configuration file"
                )

        except (FileNotFoundError, KeyError, Exception) as exp:
            # Stop execution immediately and raise the exception with a specific message
            raise Exception(f"Error : {str(exp)}")

    def verify_activation(self, section):
        """
        Verify if the plugin is activated.

        Raises:
            Exception: If the plugin is not activated.
        """
        try:
            # Call read_activation_status to check the activation status
            activation_status = self.get_config_value(self.config_file_path, section)
            # Raise an exception if the activation status is False
            if not activation_status:
                raise Exception(
                    "Plugin is not activated. Please activate the "
                    "plugin before performing this action."
                )

        except Exception as exp:
            error_message = f"{str(exp)}"
            # Log or print the error message if necessary
            print(error_message)
            raise

    def load_config(self):
        """Load configuration from the config.ini file."""
        config = configparser.ConfigParser()
        config.read(self.config_file_path)

        # Set environment variables from config file
        os.environ['API_BASEPATH'] = config.get('settings', 'API_BASEPATH',
                                                fallback='http://model-register-service/cogapi/')
        os.environ['TIMER_IN_SEC'] = config.get('settings', 'TIMER_IN_SEC', fallback='10')
        os.environ['JUPYTER_USER_ID'] = config.get('settings', 'JUPYTER_USER_ID', fallback='0')

        # Validate that the environment variables are set
        missing_vars = [var for var in ['API_BASEPATH', 'TIMER_IN_SEC', 'JUPYTER_USER_ID'] if not os.getenv(var)]
        if missing_vars:
            raise ValueError(f"The following environment variables are missing or not set: {', '.join(missing_vars)}")
